from .client import AsyncTCPClient  # noqa: F401
from .server import AsyncProtocolServer, AsyncProtocolServerClient  # noqa: F401
from .stream import AsyncTCPStream, EncryptedAsyncTCPStream  # noqa: F401



import socket
import threading
from typing import Dict, List, Tuple, Union

from pymine_net.net.server import AbstractProtocolServer, AbstractProtocolServerClient
from pymine_net.net.socket.stream import SocketTCPStream
from pymine_net.strict_abc import abstract
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket
from pymine_net.types.packet_map import PacketMap


class SocketProtocolServerClient(AbstractProtocolServerClient):
    def __init__(self, stream: SocketTCPStream, packet_map: PacketMap):
        super().__init__(stream, packet_map)
        self.stream = stream  # redefine this cause typehints

    def read_packet(self) -> ServerBoundPacket:
        length = self.stream.read_varint()
        return self._decode_packet(self.stream.read(length))

    def write_packet(self, packet: ClientBoundPacket) -> None:
        self.stream.write(self._encode_packet(packet))


class SocketProtocolServer(AbstractProtocolServer):
    def __init__(self, host: str, port: int, protocol: Union[int, str], packet_map: PacketMap):
        super().__init__(host, port, protocol, packet_map)

        self.connected_clients: Dict[Tuple[str, int], SocketProtocolServerClient] = {}

        self.sock: socket.socket = None
        self.threads: List[threading.Thread] = []
        self.running = False

    def run(self) -> None:
        self.running = True

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as self.sock:
            self.sock.bind((self.host, self.port))
            self.sock.listen(20)

            while self.running:
                connection, _ = self.sock.accept()
                self._client_connected_cb(connection)

    def close(self) -> None:
        self.sock.close()

        for thread in self.threads:
            thread.join(0.1)

    def _client_connected_cb(self, sock: socket.socket) -> None:
        client = SocketProtocolServerClient(SocketTCPStream(sock), self.packet_map)
        self.connected_clients[client.stream.remote] = client
        thread = threading.Thread(target=self._new_client_connected, args=(client,))
        self.threads.append(thread)
        thread.start()

    def _new_client_connected(self, client: SocketProtocolServerClient) -> None:
        with client.stream.sock:
            self.new_client_connected(client)

    @abstract
    def new_client_connected(self, client: SocketProtocolServerClient) -> None:
        pass



"""Contains packets related to server difficulty."""

from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket

__all__ = (
    "PlayServerDifficulty",
    "PlaySetDifficulty",
    "PlayLockDifficulty",
)


class PlayServerDifficulty(ClientBoundPacket):
    """Used by the server to update the difficulty in the client's menu. (Server -> Client)

    :param int difficulty: The difficulty level, see here: https://wiki.vg/Protocol#Server_Difficulty.
    :param bool locked: Whether the difficulty is locked or not.
    :ivar int id: Unique packet ID.
    :ivar difficulty:
    :ivar locked:
    """

    id = 0x0E

    def __init__(self, difficulty: int, locked: bool):
        super().__init__()

        self.difficulty = difficulty
        self.locked = locked

    def pack(self) -> Buffer:
        return Buffer().write("B", self.difficulty).write("?", self.locked)


class PlaySetDifficulty(ServerBoundPacket):
    """Used by the client to set difficulty. Not used normally. (Client -> Server)

    :param int new_difficulty: The new difficulty.
    :ivar int id: Unique packet ID.
    :ivar new_difficulty:
    """

    id = 0x02

    def __init__(self, new_difficulty: int) -> None:
        super().__init__()

        self.new_difficulty = new_difficulty

    @classmethod
    def unpack(cls, buf: Buffer) -> PlaySetDifficulty:
        return cls(buf.read_byte())


class PlayLockDifficulty(ServerBoundPacket):
    """Used to lock the difficulty. Only used on singleplayer. (Client -> Server)

    :param bool locked: Whether the difficulty is locked or not.
    :ivar int id: Unique packet ID.
    :ivar locked:
    """

    id = 0x10

    def __init__(self, locked: bool):
        super().__init__()

        self.locked = locked

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayLockDifficulty:
        return cls(buf.read("?"))



"""Contains packets related to bosses."""

from __future__ import annotations

from uuid import UUID

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket

__all__ = ("PlayBossBar",)


class PlayBossBar(ClientBoundPacket):
    """Used to send boss bar data. (Server -> Client)

    :param UUID uuid: UUID of the boss bar.
    :param int action: Action to take.
    :param dict **data: Data corresponding to the action.
    :ivar dict data: Data corresponding to the action.
    :ivar int id: Unique packet ID.
    :ivar uuid:
    :ivar action:
    """

    id = 0x0D

    def __init__(self, uuid: UUID, action: int, **data: dict):
        super().__init__()

        self.uuid = uuid
        self.action = action
        self.data = data

    def pack(self) -> Buffer:
        buf = Buffer.write_uuid(self.uuid).write_varint(self.action)

        if self.action == 0:
            buf.write_chat(self.data["title"]).write("f", self.data["health"]).write_varint(
                self.data["color"]
            ).write_varint(self.data["division"]).write("B", self.data["flags"])
        elif self.action == 2:
            buf.write("f", self.data["health"])
        elif self.action == 3:
            buf.write_chat(self.data["title"])
        elif self.action == 4:
            buf.write_varint(self.data["color"]).write_varint(self.data["division"])
        elif self.action == 5:
            buf.write("B", self.data["flags"])

        return buf



"""Contains packets related to beacons."""

from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ServerBoundPacket

__all__ = ("PlaySetBeaconEffect",)


class PlaySetBeaconEffect(ServerBoundPacket):
    """Changes the effect of the current beacon. (Client -> Server)

    :param int primary_effect: A potion ID (https://minecraft.gamepedia.com/Data_values#Potions).
    :param int secondary_effect: A potion ID (https://minecraft.gamepedia.com/Data_values#Potions).
    :ivar int id: Unique packet ID.
    :ivar primary_effect:
    :ivar secondary_effect:
    """

    id = 0x24

    def __init__(self, primary_effect: int, secondary_effect: int):
        super().__init__()

        self.primary_effect = primary_effect
        self.secondary_effect = secondary_effect

    @classmethod
    def unpack(cls, buf: Buffer) -> PlaySetBeaconEffect:
        return cls(buf.read_varint(), buf.read_varint())



import pymine_net.types.nbt as nbt  # noqa: F401
from pymine_net.net.asyncio.client import AsyncTCPClient  # noqa: F401
from pymine_net.packets import load_packet_map  # noqa: F401
from pymine_net.types.buffer import Buffer  # noqa: F401
from pymine_net.types.packet import ClientBoundPacket, Packet, ServerBoundPacket  # noqa: F401



from __future__ import annotations

from typing import Union


class Chat:
    def __init__(self, data: Union[str, dict]):
        if isinstance(data, str):
            self.data = {"text": data}
        else:
            self.data = data

    def __eq__(self, other: Chat) -> bool:
        return self.data == other.data



"""Contains packets related to crafting and recipes."""

from __future__ import annotations

from typing import Dict, List

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket

__all__ = (
    "PlayCraftRecipeRequest",
    "PlaySetDisplayedRecipe",
    "PlaySetRecipeBookState",
    "PlayCraftRecipeResponse",
    "PlayDeclareRecipes",
    "PlayUnlockRecipes",
)


class PlayCraftRecipeRequest(ServerBoundPacket):
    """Sent when a client/player clicks a recipe in the crafting book that is craftable. (Client -> Server)

    :param int window_id: ID of the crafting table window.
    :param str recipe_identifier: The recipe identifier.
    :param bool make_all: Whether maximum amount that can be crafted is crafted.
    :ivar int id: Unique packet ID.
    :ivar window_id:
    :ivar recipe_identifier:
    :ivar make_all:
    """

    id = 0x18

    def __init__(self, window_id: int, recipe_identifier: str, make_all: bool):
        super().__init__()

        self.window_id = window_id
        self.recipe_identifier = recipe_identifier
        self.make_all = make_all

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayCraftRecipeRequest:
        return cls(buf.read("b"), buf.read_string(), buf.read("?"))


class PlaySetDisplayedRecipe(ServerBoundPacket):
    """Replaces Recipe Book Data, type 0. See here: https://wiki.vg/Protocol#Set_Displayed_Recipe (Client -> Server)

    :param str recipe_id: The identifier for the recipe.
    :ivar int id: Unique packet ID.
    :ivar recipe_id:
    """

    id = 0x1F

    def __init__(self, recipe_id: str):
        super().__init__()

        self.recipe_id = recipe_id

    @classmethod
    def unpack(cls, buf: Buffer) -> PlaySetDisplayedRecipe:
        return cls(buf.read_string())


class PlaySetRecipeBookState(ServerBoundPacket):
    """Replaces Recipe Book Data, type 1. See here: https://wiki.vg/Protocol#Set_Recipe_Book_State (Client -> Server)

    :param int book_id: One of the following: crafting (0), furnace (1), blast furnace (2), smoker (3).
    :param bool book_open: Whether the crafting book is open or not.
    :param bool filter_active: Unknown.
    :ivar int id: Unique packet ID.
    :ivar book_id:
    :ivar book_open:
    :ivar filter_active:
    """

    id = 0x1E

    def __init__(self, book_id: int, book_open: bool, filter_active: bool):
        super().__init__()

        self.book_id = book_id
        self.book_open = book_open
        self.filter_active = filter_active

    @classmethod
    def unpack(cls, buf: Buffer) -> PlaySetRecipeBookState:
        return cls(buf.read_varint(), buf.read("?"), buf.read("?"))


class PlayCraftRecipeResponse(ClientBoundPacket):
    """Response to a PlayCraftRecipeRequest, used to update the client's UI. (Server -> Client)

    :param int window_id: ID of the crafting table window.
    :param str recipe: The recipe identifier.
    :ivar int id: Unique packet ID.
    :ivar window_id:
    :ivar recipe:
    """

    id = 0x31

    def __init__(self, window_id: int, recipe_identifier: str):
        super().__init__()

        self.window_id = window_id
        self.recipe_identifier = recipe_identifier

    def pack(self) -> Buffer:
        return Buffer().write("b", self.window_id).write_string(self.recipe_identifier)


class PlayDeclareRecipes(ClientBoundPacket):
    """Sends all registered recipes to the client. (Server -> Client)

    :param Dict[str, dict] recipes: The recipes to be sent in the form {recipe_id: recipe_data}.
    :ivar int id: Unique packet ID.
    :ivar recipes:
    """

    id = 0x66

    def __init__(self, recipes: Dict[str, dict]):
        super().__init__()

        self.recipes = recipes

    def pack(self) -> Buffer:
        buf = Buffer()

        for recipe_id, recipe in self.recipes.items():
            buf.write_recipe(recipe_id, recipe)

        return buf


class PlayUnlockRecipes(ClientBoundPacket):
    """Unlocks specified locked recipes for the client. (Server -> Client)

    :param int action: The action to be taken, see here: https://wiki.vg/Protocol#Unlock_Recipes.
    :param bool crafting_book_open: If true, then the crafting recipe book will be open when the player opens its inventory.
    :param bool crafting_book_filter_active: If true, then the filtering option is active when the players opens its inventory.
    :param bool smelting_book_open: If true, then the smelting recipe book will be open when the player opens its inventory.
    :param bool smelting_book_filter_active: If true, then the filtering option is active when the players opens its inventory.
    :param bool blast_furnace_book_open: If true, then the blast furnace recipe book will be open when the player opens its inventory.
    :param bool blast_furnace_book_filter_active: If true, then the filtering option is active when the players opens its inventory.
    :param bool smoker_book_open: If true, then the smoker recipe book will be open when the player opens its inventory.
    :param bool smoker_book_filter_active: If true, then the filtering option is active when the players opens its inventory.
    :param List[str] recipe_ids_1: First list of recipe identifiers.
    :param Optional[List[str]] recipe_ids_2: Second list of recipe identifiers.
    :ivar int id: Unique packet ID.
    :ivar action:
    :ivar crafting_book_open:
    :ivar crafting_book_filter_active:
    :ivar smelting_book_open:
    :ivar smelting_book_filter_active:
    :ivar blast_furnace_book_open:
    :ivar blast_furnace_book_filter_active:
    :ivar smoker_book_open:
    :ivar smoker_book_filter_active:
    :ivar recipe_ids_1:
    :ivar recipe_ids_2:
    """

    id = 0x39

    def __init__(
        self,
        action: int,
        crafting_book_open: bool,
        crafting_book_filter_active: bool,
        smelting_book_open: bool,
        smelting_book_filter_active: bool,
        blast_furnace_book_open: bool,
        blast_furnace_book_filter_active: bool,
        smoker_book_open: bool,
        smoker_book_filter_active: bool,
        recipe_ids_1: List[str],
        recipe_ids_2: List[list] = None,
    ):
        super().__init__()

        self.action = action
        self.crafting_book_open = crafting_book_open
        self.crafting_book_filter_active = crafting_book_filter_active
        self.smelting_book_open = smelting_book_open
        self.smelting_book_filter_active = smelting_book_filter_active
        self.blast_furnace_book_open = blast_furnace_book_open
        self.blast_furnace_book_filter_active = blast_furnace_book_filter_active
        self.smoker_book_open = smoker_book_open
        self.smoker_book_filter_active = smoker_book_filter_active
        self.recipe_ids_1 = recipe_ids_1
        self.recipe_ids_2 = recipe_ids_2

    def pack(self) -> Buffer:
        buf = (
            Buffer()
            .write_varint(self.action)
            .write("?", self.crafting_book_open)
            .write("?", self.crafting_book_filter_active)
            .write("?", self.smelting_book_open)
            .write("?", self.smelting_book_filter_active)
            .write("?", self.blast_furnace_book_open)
            .write("?", self.blast_furnace_book_filter_active)
            .write("?", self.smoker_book_open)
            .write("?", self.smoker_book_filter_active)
            .write_varint(len(self.recipe_ids_1))
        )

        for recipe_id in self.recipe_ids_1:
            buf.write_string(recipe_id)

        if self.recipe_ids_2:
            buf.write("?", True)

            for recipe_id in self.recipe_ids_2:
                buf.write_string(recipe_id)
        else:
            buf.write("?", False)

        return buf



import importlib
import os
import warnings
from typing import Dict, List, Union

from pymine_net.enums import GameState
from pymine_net.types.packet import Packet
from pymine_net.types.packet_map import DuplicatePacketIdError, PacketMap, StatePacketMap

__all__ = ("load_packet_map",)

GAME_STATES = {
    GameState.HANDSHAKING: "handshaking",
    GameState.STATUS: "status",
    GameState.LOGIN: "login",
    GameState.PLAY: "play",
}

# the directory this file is contained in
FILE_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

PROTOCOL_MAP = {757: "v_1_18_1"}


def load_packet_map(protocol: Union[int, str], *, debug: bool = False) -> PacketMap:
    protocol = PROTOCOL_MAP.get(protocol, protocol)
    packets: Dict[GameState, StatePacketMap] = {}

    for state, state_name in GAME_STATES.items():
        module_base = ["packets", str(protocol), state_name]

        packet_classes: List[Packet] = []

        # iterate through the files in the <state> folder
        for file_name in os.listdir(os.path.join(FILE_DIR, *module_base)):
            if not file_name.endswith(".py"):
                continue

            # import the file (pymine_net.packets.<protocol>.<state>.<packet group>)
            module = importlib.import_module(".".join(["pymine_net", *module_base, file_name[:-3]]))

            if debug and not hasattr(module, "__all__"):
                warnings.warn(f"{module.__name__} is missing member __all__ and cannot be loaded.")
                continue

            # iterate through object names in module.__all__
            for member_name in module.__all__:
                # attempt to get object, warn if it's not actually present
                try:
                    obj = getattr(module, member_name)
                except AttributeError:
                    if debug:
                        warnings.warn(
                            f"{module.__name__} is missing member {member_name!r} present in __all__."
                        )
                else:
                    if issubclass(obj, Packet):
                        packet_classes.append(obj)

        # attempt to create the StatePacketMap from the list of loaded packets.
        try:
            packets[state] = StatePacketMap.from_list(state, packet_classes, check_duplicates=debug)
        except DuplicatePacketIdError as e:  # re-raise with protocol included in exception
            raise DuplicatePacketIdError(protocol, e.state, e.packet_id, e.direction)

    return PacketMap(protocol, packets)



"""Contains packets related to the chat."""

from __future__ import annotations

from typing import List, Optional, Tuple
from uuid import UUID

from pymine_net.types.buffer import Buffer
from pymine_net.types.chat import Chat
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket

__all__ = (
    "PlayChatMessageClientBound",
    "PlayChatMessageServerBound",
    "PlayTabCompleteClientBound",
    "PlayTabCompleteServerBound",
)


class PlayChatMessageClientBound(ClientBoundPacket):
    """A chat message from the server to the client (Server -> Client)

    :param Chat data: The actual chat data.
    :param int position: Where on the GUI the message is to be displayed.
    :param UUID sender: Used by the Notchian client for the disableChat launch option. Setting both longs to 0 will always display the message regardless of the setting.
    :ivar int id: Unique packet ID.
    :ivar data:
    :ivar position:
    :ivar sender:
    """

    id = 0x0F

    def __init__(self, data: Chat, position: int, sender: UUID):
        super().__init__()

        self.data = data
        self.position = position
        self.sender = sender

    def pack(self) -> Buffer:
        return Buffer().write_chat(self.data).write("b", self.position).write_uuid(self.sender)


class PlayChatMessageServerBound(ServerBoundPacket):
    """A chat message from a client to the server. Can be a command. (Client -> Server)

    :param str message: The raw text sent by the client.
    :ivar int id: Unique packet ID.
    :ivar message:
    """

    id = 0x03

    def __init__(self, message: str):
        super().__init__()

        self.message = message

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayChatMessageServerBound:
        return cls(buf.read_string())


class PlayTabCompleteServerBound(ServerBoundPacket):
    """Used when a client wants to tab complete a chat message. (Client -> Server)

    :param int transaction_id: Number generated by the client.
    :param str text: All text behind / to the left of the cursor.
    :ivar int id: Unique packet ID.
    :ivar transaction_id:
    :ivar text:
    """

    id = 0x06

    def __init__(self, transaction_id: int, text: str):
        super().__init__()

        self.transaction_id = transaction_id
        self.text = text

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayTabCompleteServerBound:
        return cls(buf.read_varint(), buf.read_string())


class PlayTabCompleteClientBound(ClientBoundPacket):
    """Sends a list of auto-completions for a command. For regular chat, this would be a player username. Command names and parameters are also supported. (Server -> Client)

    :param int transaction_id: Number generated by the client.
    :param int start: Start of the text to replace.
    :param int length: Length of the text to replace.
    :param list matches: List of matches.
    :ivar int id: Unique packet ID.
    :ivar transaction_id:
    :ivar start:
    :ivar matches:
    """

    id = 0x11

    def __init__(self, transaction_id: int, start: int, matches: List[Tuple[str, Optional[Chat]]]):
        super().__init__()

        self.transaction_id = transaction_id
        self.start = start
        self.matches = matches

    def pack(self) -> Buffer:
        buf = (
            Buffer()
            .write_varint(self.id)
            .write_varint(self.start)
            .write_varint(self.length)
            .write_varint(len(self.matches))
        )

        for match, tooltip in self.matches:
            buf.write_string(match)

            if tooltip is not None:
                buf.write("?", True).write_chat(tooltip)
            else:
                buf.write("?", False)

        return buf



import socket
import struct
from typing import Tuple, Union

from cryptography.hazmat.primitives.ciphers import Cipher

from pymine_net.net.stream import AbstractTCPStream
from pymine_net.types.buffer import Buffer

__all__ = ("SocketTCPStream", "EncryptedSocketTCPStream")


class SocketTCPStream(AbstractTCPStream, socket.socket):
    """Used for reading and writing from/to a connected client, wraps a socket.socket.

    :param socket.socket sock: A socket.socket instance.
    :ivar Tuple[str, int] remote: A tuple which stores the remote client's address and port.
    :ivar sock:
    """

    __slots__ = ("sock",)

    def __init__(self, sock: socket.socket):
        self.sock = sock

        self.remote: Tuple[str, int] = sock.getsockname()

    def read(self, length: int) -> Buffer:
        result = Buffer()

        while len(result) < length:
            read_bytes = self.sock.recv(length - len(result))

            if len(read_bytes) == 0:
                raise IOError("Server didn't respond with information!")

            result.extend(read_bytes)

        return result

    def write(self, data: Union[Buffer, bytes, bytearray]) -> None:
        self.sock.sendall(data)

    def close(self) -> None:
        self.sock.close()

    def read_varint(self) -> int:
        value = 0

        for i in range(10):
            (byte,) = struct.unpack(">B", self.read(1))
            value |= (byte & 0x7F) << 7 * i

            if not byte & 0x80:
                break

        if value & (1 << 31):
            value -= 1 << 32

        value_max = (1 << (32 - 1)) - 1
        value_min = -1 << (32 - 1)

        if not (value_min <= value <= value_max):
            raise ValueError(
                f"Value doesn't fit in given range: {value_min} <= {value} < {value_max}"
            )

        return value


class EncryptedSocketTCPStream(SocketTCPStream):
    """Used for reading and writing from/to a connected client, wraps a socket.socket.

    :param socket.socket sock: A socket.socket instance.
    :param Cipher cipher: The cipher instance, used to encrypt + decrypt data.
    :ivar Tuple[str, int] remote: A tuple which stores the remote client's address and port.
    :ivar sock:
    :ivar _CipherContext decryptor: Description of parameter `_CipherContext`.
    :ivar _CipherContext encryptor: Description of parameter `_CipherContext`.
    """

    def __init__(self, sock: socket.socket, cipher: Cipher):
        super().__init__(sock)

        self.decryptor = cipher.decryptor()
        self.encryptor = cipher.encryptor()

    def read(self, length: int) -> Buffer:
        return Buffer(self.decryptor.update(super().read(length)))

    def write(self, data: Union[Buffer, bytes, bytearray]) -> None:
        super().write(self.encryptor.update(data))



import struct
from asyncio import StreamWriter
from typing import Tuple, Union

from cryptography.hazmat.primitives.ciphers import Cipher

from pymine_net.net.stream import AbstractTCPStream
from pymine_net.types.buffer import Buffer

__all__ = ("AsyncTCPStream", "EncryptedAsyncTCPStream")


class AsyncTCPStream(AbstractTCPStream, StreamWriter):
    """Used for reading and writing from/to a connected client, merges functions of a StreamReader and StreamWriter.

    :param StreamReader reader: An asyncio.StreamReader instance.
    :param StreamWriter writer: An asyncio.StreamWriter instance.
    :ivar Tuple[str, int] remote: A tuple which stores the remote client's address and port.
    """

    def __init__(self, writer: StreamWriter):
        super().__init__(writer._transport, writer._protocol, writer._reader, writer._loop)

        self.remote: Tuple[str, int] = self.get_extra_info("peername")

    async def read(self, length: int = -1) -> Buffer:
        return Buffer(await self._reader.read(length))

    async def readline(self) -> Buffer:
        return Buffer(await self._reader.readline())

    async def readexactly(self, length: int) -> Buffer:
        return Buffer(await self._reader.readexactly(length))

    async def readuntil(self, separator: bytes = b"\n") -> Buffer:
        return Buffer(await self._reader.readuntil(separator))

    async def read_varint(self) -> int:
        value = 0

        for i in range(10):
            (byte,) = struct.unpack(">B", await self.readexactly(1))
            value |= (byte & 0x7F) << 7 * i

            if not byte & 0x80:
                break

        if value & (1 << 31):
            value -= 1 << 32

        value_max = (1 << (32 - 1)) - 1
        value_min = -1 << (32 - 1)

        if not (value_min <= value <= value_max):
            raise ValueError(
                f"Value doesn't fit in given range: {value_min} <= {value} < {value_max}"
            )

        return value


class EncryptedAsyncTCPStream(AsyncTCPStream):
    """An encrypted version of an AsyncTCPStream, automatically encrypts and decrypts outgoing and incoming data.

    :param AsyncTCPStream stream: The original, stream-compatible object.
    :param Cipher cipher: The cipher instance, used to encrypt + decrypt data.
    :ivar _CipherContext decryptor: Description of parameter `_CipherContext`.
    :ivar _CipherContext encryptor: Description of parameter `_CipherContext`.
    """

    def __init__(self, stream: AsyncTCPStream, cipher: Cipher):
        super().__init__(stream)

        self.decryptor = cipher.decryptor()
        self.encryptor = cipher.encryptor()

    async def read(self, length: int = -1) -> Buffer:
        return Buffer(self.decryptor.update(await super().read(length)))

    async def readline(self) -> Buffer:
        return Buffer(self.decryptor.update(await super().readline()))

    async def readexactly(self, length: int) -> Buffer:
        return Buffer(self.decryptor.update(await super().readexactly(length)))

    async def readuntil(self, separator: bytes = b"\n") -> Buffer:
        return Buffer(self.decryptor.update(await super().readuntil(separator)))

    def write(self, data: Union[Buffer, bytes, bytearray]) -> None:
        super().write(self.encryptor.update(data))

    def writelines(self, data: Union[Buffer, bytes, bytearray]) -> None:
        super().writelines(self.encryptor.update(data))



"""Contains packets for maintaining the connection between client and server."""

from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket

__all__ = ("PlayKeepAliveClientBound", "PlayKeepAliveServerBound")


class PlayKeepAliveClientBound(ClientBoundPacket):
    """Sent by the server in order to maintain connection with the client. (Server -> Client)

    :param int keep_alive_id: A randomly generated (by the server) integer/long.
    :ivar int id: Unique packet ID.
    :ivar keep_alive_id:
    """

    id = 0x21

    def __init__(self, keep_alive_id: int):
        super().__init__()

        self.keep_alive_id = keep_alive_id

    def pack(self) -> Buffer:
        return Buffer().write("q", self.keep_alive_id)


class PlayKeepAliveServerBound(ServerBoundPacket):
    """Sent by client in order to maintain connection with server. (Client -> Server)

    :param int keep_alive_id: A randomly generated (by the server) integer/long.
    :ivar int id: Unique packet ID.
    :ivar keep_alive_id:
    """

    id = 0x0F

    def __init__(self, keep_alive_id: int):
        super().__init__()

        self.keep_alive_id = keep_alive_id

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayKeepAliveServerBound:
        return cls(buf.read("q"))



from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket

__all__ = ("StatusStatusRequest", "StatusStatusResponse", "StatusStatusPingPong")


class StatusStatusRequest(ServerBoundPacket):
    """Request from the client to get information on the server. (Client -> Server)

    :ivar int id: Unique packet ID.
    """

    id = 0x00

    def __init__(self):
        super().__init__()

    def pack(self) -> Buffer:
        return Buffer()

    @classmethod
    def unpack(cls, buf: Buffer) -> StatusStatusRequest:
        return cls()


class StatusStatusResponse(ClientBoundPacket):
    """Returns server status data back to the requesting client. (Server -> Client)

    :param dict data: JSON response data sent back to the client.
    :ivar int id: Unique packet ID.
    :ivar data:
    """

    id = 0x00

    def __init__(self, data: dict):
        super().__init__()

        self.data = data

    def pack(self) -> Buffer:
        return Buffer().write_json(self.data)

    @classmethod
    def unpack(cls, buf: Buffer) -> StatusStatusResponse:
        return cls(buf.read_json())


class StatusStatusPingPong(ServerBoundPacket, ClientBoundPacket):
    """Ping pong? (Client <-> Server)

    :param int payload: A long number, randomly generated or what the client sent.
    :ivar int id: Unique packet ID.
    :ivar int payload:
    """

    id = 0x01

    def __init__(self, payload: int):
        super().__init__()

        self.payload = payload

    @classmethod
    def unpack(cls, buf: Buffer) -> StatusStatusPingPong:
        return cls(buf.read("q"))

    def pack(self) -> Buffer:
        return Buffer().write("q", self.payload)



"""Contains the PlaySetCooldown packet."""

from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket

__all__ = ("PlaySetCooldown",)


class PlaySetCooldown(ClientBoundPacket):
    """Applies a cooldown period to all items with the given type. (Server -> Client)

    Client bound(Server -> Client)
    :param int item_id: The unique id of the type of affected items.
    :param int cooldown_ticks: The length of the cooldown in in-game ticks.
    :ivar int id: The unique ID of the packet.
    :ivar item_id:
    :ivar cooldown_ticks:
    """

    id = 0x17

    def __init__(self, item_id: int, cooldown_ticks: int) -> None:
        super().__init__()

        self.item_id = item_id
        self.cooldown_ticks = cooldown_ticks

    def pack(self) -> Buffer:
        return Buffer().write_varint(self.item_id).write_varint(self.cooldown_ticks)



"""Contains packets related to advancements."""

from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket

__all__ = ("PlayAdvancementTab", "PlaySelectAdvancementTab")


class PlayAdvancementTab(ServerBoundPacket):
    """Related to advancement tab menu, see here: https://wiki.vg/Protocol#Advancement_Tab (Client -> Server)

    :param int action: Either opened tab (0), or closed screen (1).
    :param int tab_id: The ID of the tab, only present if action is 0 (opened tab).
    :ivar int id: Unique packet ID.
    :ivar action:
    :ivar tab_id:
    """

    id = 0x22

    def __init__(self, action: int, tab_id: int):
        super().__init__()

        self.action = action
        self.tab_id = tab_id

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayAdvancementTab:
        return cls(buf.read_varint(), buf.read_optional(buf.read_varint))


class PlaySelectAdvancementTab(ClientBoundPacket):
    """Sent by the server to indicate that the client should switch advancement tab. Sent either when the client switches tab in the GUI or when an advancement in another tab is made. (Server -> Client)

    :param Optional[str] identifier: One of the following: minecraft:story/root, minecraft:nether/root, minecraft:end/root, minecraft:adventure/root, minecraft:husbandry/root.
    :ivar int id: Unique packet ID.
    :ivar Optional[str] identifier:
    """

    id = 0x40

    def __init__(self, identifier: str = None):
        super().__init__()

        self.identifier = identifier

    def pack(self) -> Buffer:
        buf = Buffer()
        return buf.write_optional(buf.write_string, self.identifier)



import zlib
from typing import Dict, Tuple, Type, Union

from pymine_net.enums import GameState, PacketDirection
from pymine_net.errors import UnknownPacketIdError
from pymine_net.net.stream import AbstractTCPStream
from pymine_net.strict_abc import StrictABC, abstract
from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket
from pymine_net.types.packet_map import PacketMap


class AbstractProtocolServerClient(StrictABC):
    __slots__ = ("stream", "packet_map", "state", "compression_threshold")

    def __init__(self, stream: AbstractTCPStream, packet_map: PacketMap):
        self.stream = stream
        self.packet_map = packet_map
        self.state = GameState.HANDSHAKING
        self.compression_threshold = -1

    def _encode_packet(self, packet: ClientBoundPacket) -> Buffer:
        """Encodes and (if necessary) compresses a ClientBoundPacket."""

        buf = Buffer().write_varint(packet.id).extend(packet.pack())

        if self.compression_threshold >= 1:
            if len(buf) >= self.compression_threshold:
                buf = Buffer().write_varint(len(buf)).extend(zlib.compress(buf))
            else:
                buf = Buffer().write_varint(0).extend(buf)

        buf = Buffer().write_varint(len(buf)).extend(buf)
        return buf

    def _decode_packet(self, buf: Buffer) -> ServerBoundPacket:
        # decompress packet if necessary
        if self.compression_threshold >= 0:
            uncompressed_length = buf.read_varint()

            if uncompressed_length > 0:
                buf = Buffer(zlib.decompress(buf.read_bytes()))

        packet_id = buf.read_varint()

        # attempt to get packet class from given state and packet id
        try:
            packet_class: Type[ClientBoundPacket] = self.packet_map[
                PacketDirection.SERVERBOUND, self.state, packet_id
            ]
        except KeyError:
            raise UnknownPacketIdError(
                self.packet_map.protocol, self.state, packet_id, PacketDirection.SERVERBOUND
            )

        return packet_class.unpack(buf)

    @abstract
    def read_packet(self) -> ServerBoundPacket:
        pass

    @abstract
    def write_packet(self, packet: ClientBoundPacket) -> None:
        pass


class AbstractProtocolServer(StrictABC):
    """Abstract class for a TCP server that handles Minecraft packets."""

    def __init__(self, host: str, port: int, protocol: Union[int, str], packet_map: PacketMap):
        self.host = host
        self.port = port
        self.protocol = protocol
        self.packet_map = packet_map

        self.connected_clients: Dict[Tuple[str, int], AbstractProtocolServerClient] = {}

    @abstract
    def run(self) -> None:
        pass

    @abstract
    def close(self) -> None:
        pass



"""Contains packets related to command blocks."""

from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ServerBoundPacket

__all__ = (
    "PlayUpdateCommandBlock",
    "PlayUpdateCommandBlockMinecart",
)


class PlayUpdateCommandBlock(ServerBoundPacket):
    """Used when a client updates a command block. (Client -> Server)

    :param int x: The x coordinate of the command block.
    :param int y: The y coordinate of the command block.
    :param int z: The z coordinate of the command block.
    :param str command: The command text in the command block.
    :param int mode: The mode which the command block is in. Either sequence (0), auto (1), or redstone (2).
    :param int flags: Other flags, see here: https://wiki.vg/Protocol#Update_Command_Block.
    :ivar int id: Unique packet ID.
    :ivar x:
    :ivar y:
    :ivar z:
    :ivar command:
    :ivar mode:
    :ivar flags:
    """

    id = 0x26

    def __init__(self, x: int, y: int, z: int, command: str, mode: int, flags: int):
        super().__init__()

        self.x, self.y, self.z = x, y, z
        self.command = command
        self.mode = mode
        self.flags = flags

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayUpdateCommandBlock:
        return cls(buf.read_position(), buf.read_string(), buf.read_varint(), buf.read("b"))


class PlayUpdateCommandBlockMinecart(ServerBoundPacket):
    """Sent when the client updates a command block minecart. (Client -> Server)

    :param int entity_id: The ID of the entity (the minecart).
    :param str command: The command text in the command block.
    :param bool track_output: Whether output from the last command is saved.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar command:
    :ivar track_output:
    """

    id = 0x27

    def __init__(self, entity_id: int, command: str, track_output: bool):
        super().__init__()

        self.entity_id = entity_id
        self.command = command
        self.track_output = track_output

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayUpdateCommandBlockMinecart:
        return cls(buf.read_varint(), buf.read_string(), buf.read("?"))



"""For packets related to sound and particle effects."""

from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket

__all__ = ("PlayEffect", "PlayEntityEffect", "PlaySoundEffect")


class PlayEffect(ClientBoundPacket):
    """Used to play a sound or particle effect. (Server -> Client)

    :param int effect_id: The ID of the effect to be played, see here: https://wiki.vg/Protocol#Effect.
    :param int x: The x coordinate where the sound/particle is played.
    :param int y: The y coordinate where the sound/particle is played.
    :param int z: The z coordinate where the sound/particle is played.
    :param int data: Extra data for certain effects.
    :param bool disable_relative_volume: If false, sound effects fade away with distance.
    :ivar int id: Unique packet ID.
    :ivar effect_id:
    :ivar x:
    :ivar y:
    :ivar z:
    :ivar data:
    :ivar disable_relative_volume:
    """

    id = 0x23

    def __init__(
        self, effect_id: int, x: int, y: int, z: int, data: int, disable_relative_volume: bool
    ):
        super().__init__()

        self.effect_id = effect_id
        self.x = x
        self.y = y
        self.z = z
        self.data = data
        self.disable_relative_volume = disable_relative_volume

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write("i", self.effect_id)
            .write_position(self.x, self.y, self.z)
            .write("i", self.data)
            .write("?", self.disable_relative_volume)
        )


class PlayEntityEffect(ClientBoundPacket):
    """Gives a specific entity an effect. (Server -> Client)

    :param int entity_id: The ID of the entity to give the effect to.
    :param int effect_id: The ID of the effect to give.
    :param int amplifier: Amplification amount of effect.
    :param int duration: The duration of the effect in ticks.
    :param int flags: Bit field, see https://wiki.vg/Protocol#Entity_Effect
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar effect_id:
    :ivar amplifier:
    :ivar duration:
    :ivar flags:
    """

    id = 0x65

    def __init__(self, entity_id: int, effect_id: int, amplifier: int, duration: int, flags: int):
        self.entity_id = entity_id
        self.effect_id = effect_id
        self.amplifier = amplifier
        self.duration = duration
        self.flags = flags

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_varint(self.entity_id)
            .write_byte(self.effect_id)
            .write_byte(self.amplifier)
            .write_varint(self.duration)
            .write_byte(self.flags)
        )


class PlaySoundEffect(ClientBoundPacket):
    """Used to play a hardcoded sound event. (Server -> Client)

    :param int sound_id: The ID of the sound to be played.
    :param int category: The sound category, see here: https://wiki.vg/Protocol#Sound_Effect.
    :param int x: The x coordinate of where the effect is to be played.
    :param int y: The y coordinate of where the effect is to be played.
    :param int z: The z coordinate of where the effect is to be played.
    :param float volume: Volume of the sound to be played, between 0.0 and 1.0.
    :param float pitch: The pitch that the sound should be played at, between 0.5 and 2.0.
    :ivar int id: Unique packet ID.
    :ivar sound_id:
    :ivar category:
    :ivar x:
    :ivar y:
    :ivar z:
    :ivar volume:
    :ivar pitch:
    """

    id = 0x5D

    def __init__(
        self, sound_id: int, category: int, x: int, y: int, z: int, volume: float, pitch: float
    ):
        super().__init__()

        self.sound_id = sound_id
        self.category = category
        self.x = x
        self.y = y
        self.z = z
        self.volume = volume
        self.pitch = pitch

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_varint(self.sound_id)
            .write_varint(self.category)
            .write("i", self.x * 8)
            .write("i", self.y * 8)
            .write("i", self.z * 8)
            .write("f", self.volume)
            .write("f", self.pitch)
        )



from strict_abc import StrictABC, abstract

__all__ = ("BlockPalette",)


class BlockPalette(StrictABC):
    @abstract
    def get_bits_per_block(self) -> int:
        pass

    @abstract
    def encode(self, block: str, props: dict = None) -> int:
        pass

    @abstract
    def decode(self, state: int) -> dict:
        pass



"""Contains packets related to commands"""

from __future__ import annotations

from typing import List

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket

__all__ = ("PlayDeclareCommands",)


class PlayDeclareCommands(ClientBoundPacket):
    """Tells the clients what commands there are. (Server -> Client)

    :param List[dict] nodes: The command nodes, a list of dictionaries. The first item is assumed to be the root node.
    :ivar int id: Unique packet ID.
    :ivar nodes:
    """

    id = 0x12

    def __init__(self, nodes: List[dict]) -> None:
        super().__init__()

        self.nodes = nodes

    def pack(self) -> Buffer:
        buf = Buffer().write_varint(len(self.nodes))

        for node in self.nodes:
            buf.write_node(node)

        return buf.write_varint(0)



from __future__ import annotations

from typing import Dict, List, Tuple, Type, Union

from pymine_net.enums import GameState, PacketDirection
from pymine_net.errors import DuplicatePacketIdError
from pymine_net.types.packet import ClientBoundPacket, Packet, ServerBoundPacket


class StatePacketMap:
    """Stores a game state's packets"""

    def __init__(
        self,
        state: GameState,
        server_bound: Dict[int, Type[ServerBoundPacket]],
        client_bound: Dict[int, Type[ClientBoundPacket]],
    ):
        self.state = state
        self.server_bound = server_bound
        self.client_bound = client_bound

    @classmethod
    def from_list(
        cls, state: GameState, packets: List[Type[Packet]], *, check_duplicates: bool = False
    ) -> StatePacketMap:
        self = cls(
            state,
            {p.id: p for p in packets if issubclass(p, ServerBoundPacket)},
            {p.id: p for p in packets if issubclass(p, ClientBoundPacket)},
        )

        if check_duplicates:
            for packet_id in self.server_bound.keys():
                found = [
                    p for p in packets if p.id == packet_id and issubclass(p, ServerBoundPacket)
                ]

                if len(found) > 1:
                    raise DuplicatePacketIdError(
                        "unknown", state, packet_id, PacketDirection.SERVERBOUND
                    )

            for packet_id in self.client_bound.keys():
                found = [
                    p for p in packets if p.id == packet_id and issubclass(p, ClientBoundPacket)
                ]

                if len(found) > 1:
                    raise DuplicatePacketIdError(
                        "unknown", state, packet_id, PacketDirection.CLIENTBOUND
                    )

        return self


class PacketMap:
    """Stores a Minecraft protocol's packets"""

    def __init__(self, protocol: Union[str, int], packets: Dict[GameState, StatePacketMap]):
        self.protocol = protocol
        self.packets = packets

    def __getitem__(self, key: Tuple[PacketDirection, int, int]) -> Packet:
        direction, state, packet_id = key

        if direction == PacketDirection.CLIENTBOUND:
            return self.packets[state].client_bound[packet_id]

        return self.packets[state].server_bound[packet_id]



from .client import SocketTCPClient  # noqa: F401
from .server import SocketProtocolServer, SocketProtocolServerClient  # noqa: F401
from .stream import EncryptedSocketTCPStream, SocketTCPStream  # noqa: F401



from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.chat import Chat
from pymine_net.types.packet import ClientBoundPacket

__all__ = ("PlayPlayerListHeaderAndFooter",)


class PlayPlayerListHeaderAndFooter(ClientBoundPacket):
    """Sent to display additional information above/below the client's player list. (Server -> Client)

    :param Chat header: Content to display above player list.
    :param Chat footer: Content to display below player list.
    :ivar int id: Unique packet ID.
    :ivar header:
    :ivar footer:
    """

    id = 0x5F

    def __init__(self, header: Chat, footer: Chat):
        super().__init__()

        self.header = header
        self.footer = footer

    def pack(self) -> Buffer:
        return Buffer().write_chat(self.header).write_chat(self.footer)



"""Contains packets related to entities."""

from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Union

import pymine_net.types.nbt as nbt
from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket

__all__ = (
    "PlayBlockEntityData",
    "PlayQueryEntityNBT",
    "PlayInteractEntity",
    "PlayEntityStatus",
    "PlayEntityAction",
    "PlayEntityPosition",
    "PlayEntityPositionAndRotation",
    "PlayEntityRotation",
    "PlayRemoveEntityEffect",
    "PlayEntityHeadLook",
    "PlayAttachEntity",
    "PlayEntityVelocity",
    "PlayEntityTeleport",
    "PlayDestroyEntities",
    "PlayEntityMetadata",
    "PlayEntityEquipment",
)


class PlayBlockEntityData(ClientBoundPacket):
    """Sets the block entity associated with the block at the given location. (Server -> Client).

    :param int x: The x coordinate of the position.
    :param int y: The y coordinate of the position.
    :param int z: The z coordinate of the position.
    :param int action: The action to be carried out (see https://wiki.vg/Protocol#Block_Entity_Data).
    :param nbt.TAG nbt_data: The nbt data associated with the action/block.
    :ivar int id: Unique packet ID.
    :ivar x:
    :ivar y:
    :ivar z:
    :ivar action:
    :ivar nbt_data:
    """

    id = 0x0A

    def __init__(self, x: int, y: int, z: int, action: int, nbt_data: nbt.TAG):
        super().__init__()

        self.x = x
        self.y = y
        self.z = z
        self.action = action
        self.nbt_data = nbt_data

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_position(self.x, self.y, self.z)
            .write("B", self.action)
            .write_nbt(self.nbt_data)
        )


class PlayQueryEntityNBT(ServerBoundPacket):
    """Sent by the client when Shift+F3+I is used. (Client -> Server)

    :param int transaction_id: Incremental ID used so the client can verify responses.
    :param int entity_id: The ID of the entity to query.
    :ivar int id: Unique packet ID.
    :ivar transaction_id:
    :ivar entity_id:
    """

    id = 0x0C

    def __init__(self, transaction_id: int, entity_id: int):
        super().__init__()

        self.transaction_id = transaction_id
        self.entity_id = entity_id

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayQueryEntityNBT:
        return cls(buf.read_varint(), buf.read_varint())


class PlayInteractEntity(ServerBoundPacket):
    """Sent when a client clicks another entity, see here: https://wiki.vg/Protocol#Interact_Entity. (Client -> Server)

    :param int entity_id: The ID of the entity interacted with.
    :param int type_: Either interact (0), attack (1), or interact at (2).
    :param Optional[int] target_x: The x coordinate of where the target is, can be None.
    :param Optional[int] target_y: The y coordinate of where the target is, can be None.
    :param Optional[int] target_z: The z coordinate of where the target is, can be None.
    :param Optional[int] hand: The hand used.
    :param bool sneaking: Whether the client was sneaking or not.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar type_:
    :ivar target_x:
    :ivar target_y:
    :ivar target_z:
    :ivar hand:
    :ivar sneaking:
    """

    id = 0x0D

    def __init__(
        self,
        entity_id: int,
        type_: int,
        target_x: Optional[int],
        target_y: Optional[int],
        target_z: Optional[int],
        hand: Optional[int],
        sneaking: bool,
    ):
        super().__init__()

        self.entity_id = entity_id
        self.type_ = type_
        self.target_x = target_x
        self.target_y = target_y
        self.target_z = target_z
        self.hand = hand
        self.sneaking = sneaking

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayInteractEntity:
        return cls(
            buf.read_varint(),
            buf.read_varint(),
            buf.read_optional(buf.read_varint),
            buf.read_optional(buf.read_varint),
            buf.read_optional(buf.read_varint),
            buf.read_optional(buf.read_varint),
            buf.read("?"),
        )


class PlayEntityStatus(ClientBoundPacket):
    """Usually used to trigger an animation for an entity. (Server -> Client)

    :param int entity_id: The ID of the entity the status is for.
    :param int entity_status: Depends on the type of entity, see here: https://wiki.vg/Protocol#Entity_Status.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar entity_status:
    """

    id = 0x1B

    def __init__(self, entity_id: int, entity_status: int):
        super().__init__()

        self.entity_id = entity_id
        self.entity_status = entity_status

    def pack(self) -> Buffer:
        return Buffer().write("i", self.entity_id).write("b", self.entity_status)


class PlayEntityAction(ServerBoundPacket):
    """Sent by the client to indicate it has performed a certain action. (Client -> Server)

    :param int entity_id: The ID of the entity.
    :param int action_id: The action occurring, see here: https://wiki.vg/Protocol#Entity_Action.
    :param int jump_boost: Used with jumping while riding a horse.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar action_id:
    :ivar jump_boost:
    """

    id = 0x1B

    def __init__(self, entity_id: int, action_id: int, jump_boost: int):
        super().__init__()

        self.entity_id = entity_id
        self.action_id = action_id
        self.jump_boost = jump_boost

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayEntityAction:
        return cls(buf.read_varint(), buf.read_varint(), buf.read_varint())


class PlayEntityPosition(ClientBoundPacket):
    """Sent by the server when an entity moves less than 8 blocks. (Server -> Client)

    :param int entity_id: The Id of the entity moving.
    :param int dx: Delta (change in) x, -8 <-> 8.
    :param int dy: Delta (change in) y, -8 <-> 8.
    :param int dz: Delta (change in) z, -8 <-> 8.
    :param bool on_ground: Whether entity is on ground or not.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar dx:
    :ivar dy:
    :ivar dz:
    :ivar on_ground:
    """

    id = 0x29

    def __init__(self, entity_id: int, dx: int, dy: int, dz: int, on_ground: bool):
        super().__init__()

        self.entity_id = entity_id
        self.dx, self.dy, self.dz = dx, dy, dz
        self.on_ground = on_ground

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_varint(self.entity_id)
            .write("h", self.dx)
            .write("h", self.dy)
            .write("h", self.dz)
            .write("?", self.on_ground)
        )


class PlayEntityPositionAndRotation(ClientBoundPacket):
    """Sent by the server when an entity rotates and moves. (Server -> Client)

    :param int entity_id: The ID of the entity moving/rotationing.
    :param int dx: Delta (change in) x, -8 <-> 8.
    :param int dy: Delta (change in) y, -8 <-> 8.
    :param int dz: Delta (change in) z, -8 <-> 8.
    :param int yaw: The new yaw angle, the value being x/256 of a full rotation.
    :param int pitch: The new pitch angle, the value being x/256 of a full rotation.
    :param bool on_ground: Whether entity is on ground or not.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar dx:
    :ivar dy:
    :ivar dz:
    :ivar yaw:
    :ivar pitch:
    :ivar on_ground:
    """

    id = 0x2A

    def __init__(
        self, entity_id: int, dx: int, dy: int, dz: int, yaw: int, pitch: int, on_ground: bool
    ):
        super().__init__()

        self.entity_id = entity_id
        self.dx, self.dy, self.dz = dx, dy, dz
        self.yaw = yaw
        self.pitch = pitch
        self.on_ground = on_ground

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_varint(self.entity_id)
            .write("h", self.dx)
            .write("h", self.dy)
            .write("h", self.dz)
            .write_byte(self.yaw)
            .write_byte(self.pitch)
            .write("?", self.on_ground)
        )


class PlayEntityRotation(ClientBoundPacket):
    """Sent by the server when an entity rotates. (Server -> Client)

    :param int entity_id: The ID of the entity.
    :param int yaw: The new yaw angle, the value being x/256 of a full rotation.
    :param int pitch: The new pitch angle, the value being x/256 of a full rotation.
    :param bool on_ground: Whether entity is on ground or not.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar yaw:
    :ivar pitch:
    :ivar on_ground:
    """

    id = 0x2B

    def __init__(self, entity_id: int, yaw: int, pitch: int, on_ground: bool):
        super().__init__()

        self.entity_id = entity_id
        self.yaw = yaw
        self.pitch = pitch
        self.on_ground = on_ground

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_varint(self.entity_id)
            .write("b", self.yaw)
            .write("b", self.pitch)
            .write("?", self.on_ground)
        )


class PlayRemoveEntityEffect(ClientBoundPacket):
    """Sent by the server to remove an entity's effect. (Server -> Client)

    :param int entity_id: The new yaw angle, the value being x/256 of a full rotation.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    """

    id = 0x3B

    def __init__(self, entity_id: int, effect_id: int):
        super().__init__()

        self.entity_id = entity_id
        self.effect_id = effect_id

    def pack(self) -> Buffer:
        return Buffer().write_varint(self.entity_id).write("b", self.effect_id)


class PlayEntityHeadLook(ClientBoundPacket):
    """Changes the horizontal direction an entity's head is facing. (Server -> Client)

    :param int entity_id: The ID of the entity.
    :param int head_yaw: The new head yaw angle, the value being x/256 of a full rotation.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar head_yaw:
    """

    id = 0x3E

    def __init__(self, entity_id: int, head_yaw: int):
        super().__init__()

        self.entity_id = entity_id
        self.head_yaw = head_yaw

    def pack(self) -> Buffer:
        return Buffer().write_varint(self.entity_id).write("B", self.head_yaw)


class PlayAttachEntity(ClientBoundPacket):
    """Sent when one entity has been leashed to another entity. (Server -> Client)

    :param int attached_entity_id: The ID of the entity attached to the leash.
    :param int holding_entity_id: The ID of the entity holding the leash.
    :ivar int id: Unique packet ID.
    :ivar attached_entity_id:
    :ivar holding_entity_id:
    """

    id = 0x4E

    def __init__(self, attached_entity_id: int, holding_entity_id: int):
        super().__init__()

        self.attached_entity_id = attached_entity_id
        self.holding_entity_id = holding_entity_id

    def pack(self) -> Buffer:
        return Buffer().write("i", self.attached_entity_id).write("i", self.holding_entity_id)


class PlayEntityVelocity(ClientBoundPacket):
    """Sends the velocity of an entity in units of 1/8000 of a block per server tick. (Server -> Client)

    :param int entity_id: The ID of the entity.
    :param int velocity_x: The velocity in units of 1/8000 of a block per server tick in the x axis.
    :param int velocity_y: The velocity in units of 1/8000 of a block per server tick in the y axis.
    :param int velocity_z: The velocity in units of 1/8000 of a block per server tick in the z axis.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar velocity_x:
    :ivar velocity_y:
    :ivar velocity_z:
    """

    id = 0x4F

    def __init__(self, entity_id: int, velocity_x: int, velocity_y: int, velocity_z: int):
        super().__init__()

        self.entity_id = entity_id
        self.velocity_x = velocity_x
        self.velocity_y = velocity_y
        self.velocity_z = velocity_z

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_varint(self.entity_id)
            .write("h", self.velocity_x)
            .write("h", self.velocity_y)
            .write("h", self.velocity_z)
        )


class PlayEntityTeleport(ClientBoundPacket):
    """Sent when an entity moves more than 8 blocks. (Server -> Client)

    :param int entity_id: The ID of the entity.
    :param float x: The new x coordinate of the entity.
    :param float y: The new y coordinate of the entity.
    :param float z: The new z coordinate of the entity.
    :param int yaw: The new yaw angle, the value being x/256 of a full rotation.
    :param int pitch: The new pitch angle, the value being x/256 of a full rotation.
    :param bool on_ground: Whether or not the entity is on the ground.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar x:
    :ivar y:
    :ivar z:
    :ivar on_ground:
    """

    id = 0x62

    def __init__(
        self, entity_id: int, x: float, y: float, z: float, yaw: int, pitch: int, on_ground: bool
    ):
        super().__init__()

        self.entity_id = entity_id
        self.x = x
        self.y = y
        self.z = z
        self.yaw = yaw
        self.pitch = pitch
        self.on_ground = on_ground

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_varint(self.entity_id)
            .write("d", self.x)
            .write("d", self.y)
            .write("d", self.z)
            .write("i", self.yaw)
            .write("i", self.pitch)
            .write("?", self.on_ground)
        )


class PlayDestroyEntities(ClientBoundPacket):
    """Sent by the server when one or more entities are to be destroyed on the client. (Server -> Client)

    :param List[int] entity_ids: List of entity IDs for the client to destroy.
    :ivar int id: Unique packet ID.
    :ivar entity_ids:
    """

    id = 0x3A

    def __init__(self, entity_ids: List[int]):
        super().__init__()

        self.entity_ids = entity_ids

    def pack(self) -> Buffer:
        buf = Buffer().write_varint(len(self.entity_ids))

        for entity_id in self.entity_ids:
            buf.write_varint(entity_id)

        return buf


class PlayEntityMetadata(ClientBoundPacket):
    """Updates one or more metadata properties for an existing entity. (Server -> Client)

    :param int entity_id: The ID of the entity the data is for.
    :param Dict[Tuple[int, int], object] metadata: The entity metadata, see here: https://wiki.vg/Protocol#Entity_Metadata.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar metadata:
    """

    id = 0x4D

    def __init__(self, entity_id: int, metadata: Dict[Tuple[int, int], object]):
        super().__init__()

        self.entity_id = entity_id
        self.metadata = metadata

    def pack(self) -> Buffer:
        return Buffer().write_varint(self.entity_id).write_entity_metadata(self.metadata)


class PlayEntityEquipment(ClientBoundPacket):
    """Sends data about the entity's equipped equipment.

    :param int entity_id: The ID of the entity the equipment data is for.
    :param List[Tuple[int, Dict[str, Union[int, nbt.TAG]]]] equipment: An array of equipment, see here: https://wiki.vg/Protocol#Entity_Equipment
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar equipment:
    """

    id = 0x50

    def __init__(self, entity_id: int, equipment: List[Tuple[int, Dict[str, Union[int, nbt.TAG]]]]):
        super().__init__()

        self.entity_id = entity_id
        self.equipment = equipment

    def pack(self) -> Buffer:
        buf = Buffer().write_varint(self.entity_id).write_varint(len(self.equipment))

        for (slot_id, equipment) in self.equipment:
            buf.write("b", slot_id).write_slot(**equipment)

        return buf


class PlayEntityProperties(ClientBoundPacket):
    """Sends information about certain attributes on an entity. (Server -> Client)

    :param int entity_id: The ID of the entity.
    :param List[dict] properties: Properties of the entity.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar properties:
    """

    id = 0x64

    def __init__(self, entity_id: int, properties: List[dict]):
        super().__init__()

        self.entity_id = entity_id
        self.properties = properties

    def pack(self) -> Buffer:
        buf = Buffer().write_varint(self.entity_id)

        for prop in self.properties:
            buf.write_string(prop["key"]).write("d", prop["value"]).write_varint(
                len(prop["modifiers"])
            )

            for prop_modifier in prop["modifiers"]:
                buf.write_modifier(prop_modifier)

        return buf



"""Contains LoginSetCompression which is technically part of the login process."""

from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket

__all__ = ("LoginSetCompression",)


class LoginSetCompression(ClientBoundPacket):
    """While not directly related to login, this packet is sent by the server during the login process. (Server -> Client)

    :param int compression_threshold: Compression level of future packets, -1 to disable compression.
    :ivar int id: Unique packet ID.
    :ivar compression_threshold:
    """

    id = 0x03

    def __init__(self, compression_threshold: int = -1):
        super().__init__()

        self.compression_threshold = compression_threshold

    def pack(self) -> Buffer:
        return Buffer().write_varint(self.compression_threshold)

    @classmethod
    def unpack(cls, buf: Buffer) -> LoginSetCompression:
        return cls(buf.read_varint())



from enum import Enum, IntEnum

__all__ = ("GameState", "Direction", "Pose", "EntityModifier")


class PacketDirection(Enum):
    CLIENTBOUND = "CLIENT-BOUND"
    SERVERBOUND = "SERVER-BOUND"


class GameState(IntEnum):
    HANDSHAKING = 0
    STATUS = 1
    LOGIN = 2
    PLAY = 3


class Direction(IntEnum):
    DOWN = 0
    UP = 1
    NORTH = 2
    SOUTH = 3
    WEST = 4
    EAST = 5


class Pose(IntEnum):
    STANDING = 0
    FALL_FLYING = 1
    SLEEPING = 2
    SWIMMING = 3
    SPIN_ATTACK = 4
    SNEAKING = 5
    DYING = 6


class EntityModifier(IntEnum):
    MODIFY = 0  # add / subtract amount
    MODIFY_PERCENT = 1  # add / subtract amount percent of the current value
    MODIFY_MULTIPLY_PERCENT = 2  # multiply by percent amount



"""Contains packets related to entities."""

from __future__ import annotations

from typing import List, Tuple

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket

__all__ = ("PlayExplosion",)


class PlayExplosion(ClientBoundPacket):
    """Sent when an explosion occurs (creepers, TNT, and ghast fireballs). (Server -> Client)

    :param float x: The x coordinate of the explosion.
    :param float y: The y coordinate of the explosion.
    :param float z: The z coordinate of the explosion.
    :param float strength: Strength of the explosion, will summon a minecraft:explosion_emitter particle if >=2.0 else a minecraft:explosion particle.
    :param List[Tuple[int, int, int]] records: Array of bytes containing the coordinates of the blocks to destroy relative to the explosion's coordinates.
    :param float pmx: Velocity to add to the player's motion in the x axis due to the explosion.
    :param float pmy: Velocity to add to the player's motion in the y axis due to the explosion.
    :param float pmz: Velocity to add to the player's motion in the z axis due to the explosion.
    :ivar int id:
    :ivar x:
    :ivar y:
    :ivar z:
    :ivar strength:
    :ivar records:
    :ivar pmx:
    :ivar pmy:
    :ivar pmz:
    """

    id = 0x1C

    def __init__(
        self,
        x: float,
        y: float,
        z: float,
        strength: float,
        records: List[Tuple[int, int, int]],
        pmx: float,
        pmy: float,
        pmz: float,
    ):
        super().__init__()

        self.x = x
        self.y = y
        self.z = z
        self.strength = strength
        self.records = records
        self.pmx = pmx
        self.pmy = pmy
        self.pmz = pmz

    def pack(self) -> Buffer:
        buf = (
            Buffer()
            .write("f", self.x)
            .write("f", self.y)
            .write("f", self.z)
            .write("f", self.strength)
            .write("i", len(self.records))
        )

        for rx, ry, rz in self.records:
            buf.write_byte(rx).write_byte(ry).write_byte(rz)

        return buf.write("f", self.pmx).write("f", self.pmy).write("f", self.pmz)



from __future__ import annotations

import gzip
import struct
from typing import List

from mutf8 import decode_modified_utf8, encode_modified_utf8

__all__ = (
    "TAG",
    "TAG_End",
    "TAG_Byte",
    "TAG_Short",
    "TAG_Int",
    "TAG_Long",
    "TAG_Float",
    "TAG_Double",
    "TAG_Byte_Array",
    "TAG_String",
    "TAG_List",
    "TAG_Compound",
    "TAG_Int_Array",
    "TAG_Long_Array",
    "TYPES",
    "unpack",
)

TYPES: List[TAG] = []


def unpack(buf, root_is_full: bool = True) -> TAG_Compound:
    try:
        data = gzip.decompress(buf)
    except gzip.BadGzipFile:
        pass
    else:
        buf.clear()
        buf.write_bytes(data)

    if root_is_full:
        buf.read_byte()
        return TAG_Compound(TAG.unpack_name(buf), TAG_Compound.unpack_data(buf))

    return TAG_Compound(None, TAG_Compound.unpack_data(buf))


class BufferUtil:
    """Reimplementation of some methods from buffer.py, needed to solve cyclic dependency issues."""

    @staticmethod
    def unpack(buf, f: str) -> object:
        unpacked = struct.unpack(">" + f, buf.read_bytes(struct.calcsize(f)))

        if len(unpacked) == 1:
            return unpacked[0]

        return unpacked

    @staticmethod
    def pack(f: str, *data: object) -> bytes:
        return struct.pack(">" + f, *data)


class TAG:
    """Base class for an NBT tag.

    :param str name: The name of the TAG.
    :ivar int id: The type ID.
    :ivar name
    """

    id = None

    def __init__(self, name: str = None):
        self.id = self.__class__.id
        self.name = "" if name is None else name

    def pack_id(self) -> bytes:
        return BufferUtil.pack("b", self.id)

    @staticmethod
    def unpack_id(buf) -> int:
        return buf.read_byte()

    def pack_name(self) -> bytes:
        mutf8_name = encode_modified_utf8(self.name)
        return BufferUtil.pack("H", len(mutf8_name)) + mutf8_name

    @staticmethod
    def unpack_name(buf) -> str:
        return decode_modified_utf8(buf.read_bytes(buf.read("H")))

    def pack_data(self) -> bytes:
        raise NotImplementedError(self.__class__.__name__)

    @classmethod
    def unpack_data(cls, buf) -> NotImplemented:
        raise NotImplementedError(cls.__name__)

    def pack(self) -> bytes:
        return self.pack_id() + self.pack_name() + self.pack_data()

    @classmethod
    def unpack(cls, buf) -> TAG:
        cls.unpack_id(buf)
        return cls(cls.unpack_name(buf), cls.unpack_data(buf))

    def pretty(self, indent: int = 0) -> str:
        return ("    " * indent) + f'{self.__class__.__name__}("{self.name}"): {self.data}'

    def __str__(self):
        return self.pretty()

    __repr__ = __str__


class TAG_End(TAG):
    id = 0

    def __init__(self, *args):
        super().__init__(None)

    def pack_name(self) -> bytes:
        return b""

    @staticmethod
    def unpack_name(buf) -> None:
        return None

    def pack_data(self) -> bytes:
        return b""

    @staticmethod
    def unpack_data(buf) -> None:
        pass

    def pretty(self, indent: int = 0) -> str:
        return ("    " * indent) + "TAG_End(): 0"


class TAG_Byte(TAG):
    """Used to represent a TAG_Byte, stores a single signed byte.

    :param str name: The name of the TAG.
    :param int data: A signed byte.
    :int id: The type ID of the TAG.
    """

    id = 1

    def __init__(self, name: str, data: int):
        super().__init__(name)

        self.data = data

    def pack_data(self) -> bytes:
        return BufferUtil.pack("b", self.data)

    @staticmethod
    def unpack_data(buf) -> int:
        return buf.read_byte()


class TAG_Short(TAG):
    """Used to represent a TAG_Short, stores a single short (2 byte int).

    :param str name: The name of the TAG.
    :param int data: A short (2 byte int).
    :int id: The type ID of the TAG.
    """

    id = 2

    def __init__(self, name: str, data: int):
        super().__init__(name)

        self.data = data

    def pack_data(self) -> bytes:
        return BufferUtil.pack("h", self.data)

    @staticmethod
    def unpack_data(buf) -> int:
        return buf.read("h")


class TAG_Int(TAG):
    """Used to represent a TAG_Int, stores an integer (4 bytes).

    :param str name: The name of the TAG.
    :param int data: A int (4 bytes).
    :int id: The type ID of the TAG.
    """

    id = 3

    def __init__(self, name: str, data: int):
        super().__init__(name)

        self.data = data

    def pack_data(self) -> bytes:
        return BufferUtil.pack("i", self.data)

    @staticmethod
    def unpack_data(buf) -> int:
        return buf.read("i")


class TAG_Long(TAG):
    """Used to represent a TAG_Long, stores a long long (8 byte int).

    :param str name: The name of the TAG.
    :param int data: A long long (8 byte int).
    :int id: The type ID of the TAG.
    """

    id = 4

    def __init__(self, name: str, data: int):
        super().__init__(name)

        self.data = data

    def pack_data(self) -> bytes:
        return BufferUtil.pack("q", self.data)

    @staticmethod
    def unpack_data(buf) -> int:
        return buf.read("q")


class TAG_Float(TAG):
    """Used to represent a TAG_Float, stores a float (4 bytes).

    :param str name: The name of the TAG.
    :param float data: A float (4 bytes).
    :int id: The type ID of the TAG.
    """

    id = 5

    def __init__(self, name: str, data: float):
        super().__init__(name)

        self.data = data

    def pack_data(self) -> bytes:
        return BufferUtil.pack("f", self.data)

    @staticmethod
    def unpack_data(buf) -> float:
        return buf.read("f")


class TAG_Double(TAG):
    """Used to represent a TAG_Double, stores a double (8 byte float).

    :param str name: The name of the TAG.
    :param float data: A double (8 byte float).
    :int id: The type ID of the TAG.
    """

    id = 6

    def __init__(self, name: str, data: float):
        super().__init__(name)

        self.data = data

    def pack_data(self) -> bytes:
        return BufferUtil.pack("d", self.data)

    @staticmethod
    def unpack_data(buf) -> float:
        return buf.read("d")


class TAG_Byte_Array(TAG, bytearray):
    """Used to represent a TAG_Byte_Array, stores an array of bytes.

    :param str name: The name of the TAG.
    :param bytearray data: Some bytes.
    :int id: The type ID of the TAG.
    """

    id = 7

    def __init__(self, name: str, data: bytearray):
        TAG.__init__(self, name)

        if isinstance(data, str):
            print(f"WARNING: data passed was not bytes ({repr(data)})")
            data = data.encode("utf8")

        bytearray.__init__(self, data)

    def pack_data(self) -> bytes:
        return BufferUtil.pack("i", len(self)) + bytes(self)

    @staticmethod
    def unpack_data(buf) -> bytearray:
        return bytearray(buf.read_bytes(buf.read("i")))

    def pretty(self, indent: int = 0) -> str:
        return f'{" " * 4 * indent}TAG_Byte_Array("{self.name}"): [{", ".join([str(v) for v in self])}]'


class TAG_String(TAG):
    """Used to represent a TAG_String, stores a string.

    :param str name: The name of the TAG.
    :param str data: A string.
    :int id: The type ID of the TAG.
    """

    id = 8

    def __init__(self, name: str, data: str):
        super().__init__(name)

        self.data = data

    def pack_data(self) -> bytes:
        mutf8_text = encode_modified_utf8(self.data)
        return BufferUtil.pack("H", len(mutf8_text)) + mutf8_text

    @staticmethod
    def unpack_data(buf) -> str:
        return decode_modified_utf8(buf.read_bytes(buf.read("H")))

    def pretty(self, indent: int = 0) -> str:
        return f'{" " * 4 * indent}{self.__class__.__name__}("{self.name}"): {self.data}'


class TAG_List(TAG, list):
    """Represents a TAG_List, a list of nameless and typeless tagss.

    :param str name: The name of the TAG.
    :param list data: A uniform list of TAGs.
    :int id: The type ID of the TAG.
    """

    id = 9

    def __init__(self, name: str, data: List[TAG]):
        TAG.__init__(self, name)
        list.__init__(self, data)

    def pack_data(self) -> bytes:
        if len(self) > 0:
            return (
                BufferUtil.pack("b", self[0].id)
                + BufferUtil.pack("i", len(self))
                + b"".join([t.pack_data() for t in self])
            )

        return BufferUtil.pack("b", 0) + BufferUtil.pack("i", 0)

    @staticmethod
    def unpack_data(buf) -> List[TAG]:
        tag = TYPES[buf.read("b")]
        length = buf.read("i")

        return [tag(None, tag.unpack_data(buf)) for _ in range(length)]

    def pretty(self, indent: int = 0) -> str:
        tab = " " * 4 * indent
        nl = ",\n"

        return f'TAG_List("{self.name}"): [\n{nl.join([t.pretty(indent+1) for t in self])}\n{tab}]'


class TAG_Compound(TAG, dict):
    """Represents a TAG_Compound, a list of named tags.

    :param str name: The name of the TAG.
    :param list data: A list of tags.
    :int id: The type ID of the TAG.
    """

    id = 10

    def __init__(self, name: str, data: List[TAG]):
        TAG.__init__(self, name)
        dict.__init__(self, [(t.name, t) for t in data])

    @property
    def data(self):
        return self.values()

    def __setitem__(self, key: str, value: TAG):
        value.name = key
        dict.__setitem__(self, key, value)

    def update(self, *args, **kwargs):
        dict.update(self, *args, **kwargs)

        for k, v in self.items():
            v.name = k

    def pack_data(self) -> bytes:
        return b"".join([tag.pack() for tag in self.values()]) + b"\x00"

    @staticmethod
    def unpack_data(buf) -> List[TAG]:
        out = []

        while True:
            tag = TYPES[buf.read("b")]

            if tag is TAG_End:
                break

            out.append(tag(tag.unpack_name(buf), tag.unpack_data(buf)))

        return out

    def pretty(self, indent: int = 0) -> str:
        tab = " " * 4 * indent
        nl = ",\n"

        return f'{tab}TAG_Compound("{self.name}"): [\n{nl.join([t.pretty(indent + 1) for t in self.values()])}\n{tab}]'


class TAG_Int_Array(TAG, list):
    """Represents a TAG_Int_Array, a list of ints (4 bytes each).

    :param str name: The name of the TAG.
    :param list data: A list of ints (4 bytes each).
    :int id: The type ID of the TAG.
    """

    id = 11

    def __init__(self, name: str, data: list):
        TAG.__init__(self, name)
        list.__init__(self, data)

    def pack_data(self) -> bytes:
        return BufferUtil.pack("i", len(self)) + b"".join(
            [BufferUtil.pack("i", num) for num in self]
        )

    @staticmethod
    def unpack_data(buf) -> List[int]:
        return [buf.read("i") for _ in range(buf.read("i"))]

    def pretty(self, indent: int = 0) -> str:
        return (
            f'{" " * 4 * indent}TAG_Int_Array("{self.name}"): [{", ".join([str(v) for v in self])}]'
        )


class TAG_Long_Array(TAG, list):
    """Represents a TAG_Long_Array, a list of long longs (8 byte ints).

    :param str name: The name of the TAG.
    :param list value: A list of long longs (8 byte ints).
    :int id: The type ID of the TAG.
    """

    id = 12

    def __init__(self, name: str, data: List[int]):
        TAG.__init__(self, name)
        list.__init__(self, data)

    def pack_data(self) -> bytes:
        return BufferUtil.pack("i", len(self)) + b"".join(
            [BufferUtil.pack("q", num) for num in self]
        )

    @staticmethod
    def unpack_data(buf) -> list:
        return [buf.read("q") for _ in range(buf.read("i"))]

    def pretty(self, indent: int = 0) -> str:
        return f'{" " * 4 * indent}TAG_Long_Array("{self.name}"): [{", ".join([str(v) for v in self])}]'


TYPES.extend(
    [
        TAG_End,
        TAG_Byte,
        TAG_Short,
        TAG_Int,
        TAG_Long,
        TAG_Float,
        TAG_Double,
        TAG_Byte_Array,
        TAG_String,
        TAG_List,
        TAG_Compound,
        TAG_Int_Array,
        TAG_Long_Array,
    ]
)



from __future__ import annotations

import json
import struct
import uuid
from typing import Callable, Dict, Optional, Tuple, Union

from pymine_net.enums import Direction, EntityModifier, Pose
from pymine_net.types import nbt
from pymine_net.types.chat import Chat
from pymine_net.types.registry import Registry

__all__ = ("Buffer",)


class Buffer(bytearray):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pos = 0

    def write_bytes(self, data: Union[bytes, bytearray]) -> Buffer:
        """Writes bytes to the buffer."""

        return self.extend(data)

    def read_bytes(self, length: int = None) -> bytearray:
        """Reads bytes from the buffer, if length is None then all bytes are read."""

        if length is None:
            length = len(self)

        try:
            return self[self.pos : self.pos + length]
        finally:
            self.pos += length

    def clear(self) -> None:
        """Resets the position and clears the bytearray."""

        super().clear()
        self.pos = 0

    def reset(self) -> None:
        """Resets the position in the buffer."""

        self.pos = 0

    def extend(self, data: Union[Buffer, bytes, bytearray]) -> Buffer:
        super().extend(data)
        return self

    def read_byte(self) -> int:
        """Reads a singular byte as an integer from the buffer."""

        byte = self[self.pos]
        self.pos += 1
        return byte

    def write_byte(self, value: int) -> Buffer:
        """Writes a singular byte to the buffer."""

        return self.extend(struct.pack(">b", value))

    def read(self, fmt: str) -> Union[object, Tuple[object]]:
        """Using the given format, reads from the buffer and returns the unpacked value."""

        unpacked = struct.unpack(">" + fmt, self.read_bytes(struct.calcsize(fmt)))

        if len(unpacked) == 1:
            return unpacked[0]

        return unpacked

    def write(self, fmt: str, *value: object) -> Buffer:
        """Using the given format and value, packs the value and writes it to the buffer."""

        self.write_bytes(struct.pack(">" + fmt, *value))
        return self

    def read_optional(self, reader: Callable) -> Optional[object]:
        """Reads an optional value from the buffer."""

        if self.read("?"):
            return reader()

    def write_optional(self, writer: Callable, value: object = None) -> Buffer:
        """Writes an optional value to the buffer."""

        if value is None:
            self.write("?", False)
        else:
            self.write("?", True)
            writer(value)

        return self

    def read_varint(self, max_bits: int = 32) -> int:
        """Reads a varint from the buffer."""

        value = 0

        for i in range(10):
            byte = self.read("B")
            value |= (byte & 0x7F) << 7 * i

            if not byte & 0x80:
                break

        if value & (1 << 31):
            value -= 1 << 32

        value_max = (1 << (max_bits - 1)) - 1
        value_min = -1 << (max_bits - 1)

        if not (value_min <= value <= value_max):
            raise ValueError(
                f"Value doesn't fit in given range: {value_min} <= {value} < {value_max}"
            )

        return value

    def write_varint(self, value: int, max_bits: int = 32) -> Buffer:
        """Writes a varint to the buffer."""

        value_max = (1 << (max_bits - 1)) - 1
        value_min = -1 << (max_bits - 1)

        if not (value_min <= value <= value_max):
            raise ValueError(
                f"num doesn't fit in given range: {value_min} <= {value} < {value_max}"
            )

        if value < 0:
            value += 1 << 32

        for _ in range(10):
            byte = value & 0x7F
            value >>= 7

            self.write("B", byte | (0x80 if value > 0 else 0))

            if value == 0:
                break

        return self

    def read_optional_varint(self) -> Optional[int]:
        """Reads an optional (None if not present) varint from the buffer."""

        value = self.read_varint()

        if value == 0:
            return None

        return value - 1

    def write_optional_varint(self, value: int = None) -> Buffer:
        """Writes an optional (None if not present) varint to the buffer."""

        return self.write_varint(0 if value is None else value + 1)

    def read_string(self) -> str:
        """Reads a UTF8 string from the buffer."""

        return self.read_bytes(self.read_varint(max_bits=16)).decode("utf-8")

    def write_string(self, value: str) -> Buffer:
        """Writes a string in UTF8 to the buffer."""

        encoded = value.encode("utf-8")
        self.write_varint(len(encoded), max_bits=16).write_bytes(encoded)

        return self

    def read_json(self) -> object:
        """Reads json data from the buffer."""

        return json.loads(self.read_string())

    def write_json(self, value: object) -> Buffer:
        """Writes json data to the buffer."""

        return self.write_string(json.dumps(value))

    def read_nbt(self) -> nbt.TAG_Compound:
        """Reads an nbt tag from the buffer."""

        return nbt.unpack(self[self.pos :])

    def write_nbt(self, value: nbt.TAG = None) -> Buffer:
        """Writes an nbt tag to the buffer."""

        if value is None:
            self.write_byte(0)
        else:
            self.write_bytes(value.pack())

        return self

    def read_uuid(self) -> uuid.UUID:
        """Reads a UUID from the buffer."""

        return uuid.UUID(bytes=bytes(self.read_bytes(16)))

    def write_uuid(self, value: uuid.UUID) -> Buffer:
        """Writes a UUID to the buffer."""

        return self.write_bytes(value.bytes)

    def read_position(self) -> Tuple[int, int, int]:
        """Reads a Minecraft position (x, y, z) from the buffer."""

        def from_twos_complement(num, bits):
            if num & (1 << (bits - 1)) != 0:
                num -= 1 << bits

            return num

        data = self.read("Q")

        return (
            from_twos_complement(data >> 38, 26),
            from_twos_complement(data & 0xFFF, 12),
            from_twos_complement(data >> 12 & 0x3FFFFFF, 26),
        )

    def read_chat(self) -> Chat:
        """Reads a chat message from the buffer."""

        return Chat(self.read_json())

    def write_chat(self, value: Chat) -> Buffer:
        """Writes a chat message to the buffer."""

        return self.write_json(value.data)

    def write_position(self, x: int, y: int, z: int) -> Buffer:
        """Writes a Minecraft position (x, y, z) to the buffer."""

        def to_twos_complement(num, bits):
            return num + (1 << bits) if num < 0 else num

        return self.write(
            "Q",
            to_twos_complement(x, 26)
            << 38 + to_twos_complement(z, 26)
            << 12 + to_twos_complement(y, 12),
        )

    def read_slot(self, registry: Registry) -> dict:
        """Reads an inventory / container slot from the buffer."""

        has_item_id = self.read_optional(self.read_varint)

        if has_item_id is None:
            return {"item": None}

        return {
            "item": registry.decode(self.read_varint()),
            "count": self.read("b"),
            "tag": self.read_nbt(),
        }

    def write_slot(self, item_id: int = None, count: int = 1, tag: nbt.TAG = None) -> Buffer:
        """Writes an inventory / container slot to the buffer."""

        if item_id is None:
            self.write("?", False)
        else:
            self.write("?", True).write_varint(item_id).write("b", count).write_nbt(tag)

    def read_rotation(self) -> Tuple[float, float, float]:
        """Reads a rotation from the buffer."""

        return self.read("fff")

    def write_rotation(self, x: float, y: float, z: float) -> Buffer:
        """Writes a rotation to the buffer."""

        return self.write("fff", x, y, z)

    def read_direction(self) -> Direction:
        """Reads a direction from the buffer."""

        return Direction(self.read_varint())

    def write_direction(self, value: Direction) -> Buffer:
        """Writes a direction to the buffer."""

        return self.write_varint(value.value)

    def read_pose(self) -> Pose:
        """Reads a pose from the buffer."""

        return Pose(self.read_varint())

    def write_pose(self, value: Pose) -> Buffer:
        """Writes a pose to the buffer."""

        return self.write_varint(value.value)

    def write_recipe_item(self, value: Union[dict, str]) -> Buffer:
        """Writes a recipe item / slot to the buffer."""

        if isinstance(value, dict):
            self.write_slot(**value)
        elif isinstance(value, str):
            self.write_slot(value)
        else:
            raise TypeError(f"Invalid type {type(value)}.")

        return self

    def write_ingredient(self, value: dict) -> Buffer:
        """Writes a part of a recipe to the buffer."""

        self.write_varint(len(value))

        for slot in value.values():
            self.write_recipe_item(slot)

    def write_recipe(self, recipe_id: str, recipe: dict) -> Buffer:
        """Writes a recipe to the buffer."""

        recipe_type = recipe["type"]

        self.write_string(recipe_type).write_string(recipe_id)

        if recipe.get("group") is None:
            recipe["group"] = "null"

        if recipe_type == "minecraft:crafting_shapeless":
            print(recipe.get("ingredients"))

            self.write_string(recipe["group"]).write_varint(len(recipe["ingredients"]))

            for ingredient in recipe.get("ingredients", []):
                self.write_ingredient(ingredient)

            self.write_recipe_item(recipe["result"])
        elif recipe_type == "minecraft:crafting_shaped":
            self.write_varint(len(recipe["patern"][0])).write_varint(
                len(recipe["pattern"])
            ).write_string(recipe["group"])

            for ingredient in recipe.get("ingredients", []):
                self.write_ingredient(ingredient)

            self.write_recipe_item(recipe["result"])
        elif recipe_type[10:] in ("smelting", "blasting", "campfire_cooking"):
            print(recipe)

            (
                self.write_string(recipe["group"])
                .write_ingredient(recipe["ingredient"])
                .write_recipe_item(recipe["result"])
                .write("f", recipe["experience"])
                .write(recipe["cookingtime"])
            )
        elif recipe_type == "minecraft:stonecutting":
            self.write_string(recipe["group"]).write_ingredient(
                recipe["ingredient"]
            ).write_recipe_item(recipe["result"])
        elif recipe_type == "minecraft:smithing":
            self.write_ingredient(recipe["base"]).write_ingredient(
                recipe["addition"]
            ).write_ingredient(recipe["result"])

        return self

    def read_villager(self) -> dict:
        """Reads villager data from the buffer."""

        return {
            "kind": self.read_varint(),
            "profession": self.read_varint(),
            "level": self.read_varint(),
        }

    def write_villager(self, kind: int, profession: int, level: int) -> Buffer:
        return self.write_varint(kind).write_varint(profession).write_varint(level)

    def write_trade(
        self,
        in_item_1: dict,
        out_item: dict,
        disabled: bool,
        num_trade_usages: int,
        max_trade_usages: int,
        xp: int,
        special_price: int,
        price_multi: float,
        demand: int,
        in_item_2: dict = None,
    ) -> Buffer:
        self.write_slot(**in_item_1).write_slot(**out_item)

        if in_item_2 is not None:
            self.write("?", True).write_slot(**in_item_2)
        else:
            self.write("?", False)

        return (
            self.write("?", disabled)
            .write("i", num_trade_usages)
            .write("i", max_trade_usages)
            .write("i", xp)
            .write("i", special_price)
            .write("f", price_multi)
            .write("i", demand)
        )

    def read_particle(self) -> dict:
        particle = {}
        particle_id = particle["id"] = self.read_varint()

        if particle_id == 3 or particle_id == 23:
            particle["block_state"] = self.read_varint()
        elif particle_id == 14:
            particle["red"] = self.read("f")
            particle["green"] = self.read("f")
            particle["blue"] = self.read("f")
            particle["scale"] = self.read("f")
        elif particle_id == 32:
            particle["item"] = self.read_slot()

        return particle

    def write_particle(self, **value) -> Buffer:
        particle_id = value["particle_id"]

        if particle_id == 3 or particle_id == 23:
            self.write_varint(value["block_state"])
        elif particle_id == 14:
            self.write("ffff", value["red"], value["green"], value["blue"], value["scale"])
        elif particle_id == 32:
            self.write_slot(**value["item"])

        return self

    def write_entity_metadata(self, value: Dict[Tuple[int, int], object]) -> Buffer:
        #  index, type, value
        for (i, t), v in value.items():
            self.write("B", i).write_varint(t)

            if t == 0:
                self.write("b", v)
            elif t == 1:
                self.write_varint(v)
            elif t == 2:
                self.write("f", v)
            elif t == 3:
                self.write_string(v)
            elif t == 4:
                self.write_chat(v)
            elif t == 5:
                self.write_optional(self.write_chat, v)
            elif t == 6:
                self.write_slot(**v)
            elif t == 7:
                self.write("?", v)
            elif t == 8:
                self.write_rotation(*v)
            elif t == 9:
                self.write_position(*v)
            elif t == 10:
                self.write("?", v is not None)
                if v is not None:
                    self.write_position(*v)
            elif t == 11:
                self.write_direction(v)
            elif t == 12:
                self.write_optional(self.write_uuid, v)
            elif t == 13:
                self.write_block(v)
            elif t == 14:
                self.write_nbt(v)
            elif t == 15:
                self.write_particle(**v)
            elif t == 16:
                self.write_villager(*v)
            elif t == 17:
                self.write_optional_varint(v)
            elif t == 18:
                self.write_pose(v)

        self.write_bytes(b"\xFE")

        return self

    def read_modifier(self) -> Tuple[uuid.UUID, float, EntityModifier]:
        return (self.read_uuid(), self.read("f"), EntityModifier(self.read("b")))

    def write_modifier(self, uuid_: uuid.UUID, amount: float, operation: EntityModifier):
        return self.write_uuid(uuid_).write("f", amount).write("b", operation)

    def write_node(self, node: dict) -> Buffer:
        node_flags = node["flags"]
        self.write_byte(node_flags).write_varint(len(node["children"]))

        for child in node["children"]:
            self.write_node(child)

        if node_flags & 0x08:
            self.write_varint(node["redirect_node"])

        if 1 >= node_flags & 0x03 <= 2:  # argument or literal node
            self.write_string(node["name"])

        if node_flags & 0x3 == 2:  # argument node
            self.write_string(node["parser"])

            if node.get("properties"):
                for writer, data in node["properties"]:
                    writer(data)

        if node_flags & 0x10:
            self.write_string(node["suggestions_type"])

        return self



import socket
from typing import Union

from pymine_net.net.client import AbstractTCPClient
from pymine_net.net.socket.stream import SocketTCPStream
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket
from pymine_net.types.packet_map import PacketMap

__all__ = ("SocketTCPClient",)


class SocketTCPClient(AbstractTCPClient):
    """A connection over a TCP socket for reading + writing Minecraft packets."""

    def __init__(self, host: str, port: int, protocol: Union[int, str], packet_map: PacketMap):
        super().__init__(host, port, protocol, packet_map)

        self.stream: SocketTCPStream = None

    def connect(self) -> None:
        sock = socket.create_connection((self.host, self.port))
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

        self.stream = SocketTCPStream(sock)

    def close(self) -> None:
        self.stream.close()

    def read_packet(self) -> ClientBoundPacket:
        packet_length = self.stream.read_varint()
        return self._decode_packet(self.stream.read(packet_length))

    def write_packet(self, packet: ServerBoundPacket) -> None:
        self.stream.write(self._encode_packet(packet))



from typing import Union

from pymine_net.enums import GameState, PacketDirection


class PyMineNetError(Exception):
    pass


class UnknownPacketIdError(Exception):
    def __init__(
        self,
        protocol: Union[str, int],
        state: GameState,
        packet_id: int,
        direction: PacketDirection,
    ):
        super().__init__(
            f"Unknown packet ID 0x{packet_id:02X} (protocol={protocol}, state={state.name}, {direction.value})"
        )

        self.protocol = protocol
        self.state = state
        self.packet_id = packet_id
        self.direction = direction


class DuplicatePacketIdError(Exception):
    def __init__(
        self,
        protocol: Union[str, int],
        state: GameState,
        packet_id: int,
        direction: PacketDirection,
    ):
        super().__init__(
            f"Duplicate packet ID found (protocol={protocol}, state={state.name}, {direction}): 0x{packet_id:02X}"
        )

        self.protocol = protocol
        self.state = state
        self.packet_id = packet_id
        self.direction = direction



"""Contains packets relating to client logins."""

from __future__ import annotations

from uuid import UUID

from pymine_net.types.buffer import Buffer
from pymine_net.types.chat import Chat
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket

__all__ = (
    "LoginStart",
    "LoginEncryptionRequest",
    "LoginEncryptionResponse",
    "LoginSuccess",
    "LoginDisconnect",
)


class LoginStart(ServerBoundPacket):
    """Packet from client asking to start login process. (Client -> Server)

    :param str username: Username of the client who sent the request.
    :ivar int id: Unique packet ID.
    :ivar username:
    """

    id = 0x00

    def __init__(self, username: str):
        super().__init__()

        self.username = username

    def pack(self) -> Buffer:
        return Buffer().write_string(self.username)

    @classmethod
    def unpack(cls, buf: Buffer) -> LoginStart:
        return cls(buf.read_string())


class LoginEncryptionRequest(ClientBoundPacket):
    """Used by the server to ask the client to encrypt the login process. (Server -> Client)

    :param bytes public_key: Public key.
    :param bytes verify_token: Verify token.
    :ivar int id: Unique packet ID.
    :ivar public_key:
    """

    id = 0x01

    def __init__(self, public_key: bytes, verify_token: bytes):
        super().__init__()

        self.public_key = public_key
        self.verify_token = verify_token

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_string(" " * 20)
            .write_varint(len(self.public_key))
            .write_bytes(self.public_key)
            .write_varint(len(self.verify_token))
            .write_bytes(self.verify_token)
        )

    @classmethod
    def unpack(cls, buf: Buffer) -> LoginEncryptionRequest:
        buf.read_string()

        return cls(buf.read_bytes(buf.read_varint()), buf.read_bytes(buf.read_varint()))


class LoginEncryptionResponse(ServerBoundPacket):
    """Response from the client to a LoginEncryptionRequest. (Client -> Server)

    :param bytes shared_key: The shared key used in the login process.
    :param bytes verify_token: The verify token used in the login process.
    :ivar int id: Unique packet ID.
    :ivar shared_key:
    :ivar verify_token:
    """

    id = 0x01

    def __init__(self, shared_key: bytes, verify_token: bytes):
        super().__init__()

        self.shared_key = shared_key
        self.verify_token = verify_token

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_varint(len(self.shared_key))
            .write_bytes(self.shared_key)
            .write_varint(len(self.verify_token))
            .write_bytes(self.verify_token)
        )

    @classmethod
    def unpack(cls, buf: Buffer) -> LoginEncryptionResponse:
        return cls(buf.read_bytes(buf.read_varint()), buf.read_bytes(buf.read_varint()))


class LoginSuccess(ClientBoundPacket):
    """Sent by the server to denote a successfull login. (Server -> Client)

    :param UUID uuid: The UUID of the connecting player/client.
    :param str username: The username of the connecting player/client.
    :ivar int id: Unique packet ID.
    :ivar uuid:
    :ivar username:
    """

    id = 0x02

    def __init__(self, uuid: UUID, username: str):
        super().__init__()

        self.uuid = uuid
        self.username = username

    def pack(self) -> Buffer:
        return Buffer().write_uuid(self.uuid).write_string(self.username)

    @classmethod
    def unpack(cls, buf: Buffer) -> LoginSuccess:
        return cls(buf.read_uuid(), buf.read_string())


class LoginDisconnect(ClientBoundPacket):
    """Sent by the server to kick a player while in the login state. (Server -> Client)

    :param Chat reason: The reason for the disconnect.
    :ivar int id: Unique packet ID.
    """

    id = 0x00

    def __init__(self, reason: Chat):
        super().__init__()

        self.reason = reason

    def pack(self) -> Buffer:
        return Buffer().write_chat(self.reason)

    @classmethod
    def unpack(cls, buf: Buffer) -> LoginDisconnect:
        return cls(buf.read_chat())


class LoginPluginRequest(ClientBoundPacket):
    """Sent by server to implement a custom handshaking flow.

    :param int message_id: Message id, generated by the server, should be unique to the connection.
    :param str channel: Channel identifier, name of the plugin channel used to send the data.
    :param bytes data: Data that is to be sent.
    :ivar int id: Unique packet ID.
    """

    id = 0x04

    def __init__(self, message_id: int, channel: str, data: bytes):
        self.message_id = message_id
        self.channel = channel
        self.data = data

    def pack(self) -> Buffer:
        return (
            Buffer().write_varint(self.message_id).write_string(self.channel).write_bytes(self.data)
        )

    @classmethod
    def unpack(cls, buf: Buffer) -> LoginPluginRequest:
        return cls(buf.read_varint(), buf.read_string(), buf.read_bytes())


class LoginPluginResponse(ServerBoundPacket):
    """Response to LoginPluginRequest from client.

    :param int message_id: Message id, generated by the server, should be unique to the connection.
    :param Optional[bytes] data: Optional response data, present if client understood request.
    :ivar int id: Unique packet ID.
    """

    id = 0x02

    def __init__(self, message_id: int, data: bytes = None):
        self.message_id = message_id
        self.data = data

    def pack(self) -> Buffer:
        buf = Buffer().write_varint(self.message_id)
        return buf.write_optional(buf.write_bytes, self.data)

    @classmethod
    def unpack(cls, buf: Buffer) -> LoginPluginResponse:
        return cls(buf.read_varint(), buf.read_optional(buf.read_bytes))



import zlib
from typing import Type, Union

from pymine_net.enums import GameState, PacketDirection
from pymine_net.errors import UnknownPacketIdError
from pymine_net.strict_abc import StrictABC, abstract
from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket
from pymine_net.types.packet_map import PacketMap


class AbstractTCPClient(StrictABC):
    """Abstract class for a connection over a TCP socket for reading + writing Minecraft packets."""

    def __init__(self, host: str, port: int, protocol: Union[int, str], packet_map: PacketMap):
        self.host = host
        self.port = port
        self.protocol = protocol
        self.packet_map = packet_map

        self.state = GameState.HANDSHAKING
        self.compression_threshold = -1

    @abstract
    def connect(self) -> None:
        pass

    @abstract
    def close(self) -> None:
        pass

    @staticmethod
    def _encode_packet(packet: ServerBoundPacket, compression_threshold: int = -1) -> Buffer:
        """Encodes and (if necessary) compresses a ServerBoundPacket."""

        buf = Buffer().write_varint(packet.id).extend(packet.pack())

        if compression_threshold >= 1:
            if len(buf) >= compression_threshold:
                buf = Buffer().write_varint(len(buf)).extend(zlib.compress(buf))
            else:
                buf = Buffer().write_varint(0).extend(buf)

        return Buffer().write_varint(len(buf)).extend(buf)

    def _decode_packet(self, buf: Buffer) -> ClientBoundPacket:
        # decompress packet if necessary
        if self.compression_threshold >= 0:
            uncompressed_length = buf.read_varint()

            if uncompressed_length > 0:
                buf = Buffer(zlib.decompress(buf.read_bytes()))

        packet_id = buf.read_varint()

        # attempt to get packet class from given state and packet id
        try:
            packet_class: Type[ClientBoundPacket] = self.packet_map[
                PacketDirection.CLIENTBOUND, self.state, packet_id
            ]
        except KeyError:
            raise UnknownPacketIdError(None, self.state, packet_id, PacketDirection.CLIENTBOUND)

        return packet_class.unpack(buf)

    @abstract
    def read_packet(self) -> ClientBoundPacket:
        pass

    @abstract
    def write_packet(self, packet: ServerBoundPacket) -> None:
        pass



"""Contains animation packets."""

from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket

__all__ = (
    "PlayEntityAnimation",
    "PlayBlockBreakAnimation",
    "PlayAnimationServerBound",
    "PlayOpenBook",
)


class PlayEntityAnimation(ClientBoundPacket):
    """Sent whenever an entity should change animation. (Server -> Client)

    :param int entity_id: Entity ID of the digging entity.
    :param int animation: Value 0-5 which correspond to a specific animation (https://wiki.vg/Protocol#Entity_Animation_.28clientbound.29).
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar animation:
    """

    id = 0x06

    def __init__(self, entity_id: int, animation: int):
        super().__init__()

        self.entity_id = entity_id
        self.animation = animation

    def pack(self) -> Buffer:
        return Buffer().write_varint(self.entity_id).write("B", self.animation)


class PlayBlockBreakAnimation(ClientBoundPacket):
    """Sent to play a block breaking animation. (Server -> Client)

    :param int entity_id: Entity ID of the entity which broke the block, or random.
    :param int x: The x coordinate of the location to play the animation.
    :param int y: The y coordinate of the location to play the animation.
    :param int z: The z coordinate of the location to play the animation.
    :param int stage: Stage from 0-9 in the breaking animation.
    :ivar int id: Unique packet ID.
    :ivar entity_id:
    :ivar x:
    :ivar y:
    :ivar z:
    :ivar stage:
    """

    id = 0x09

    def __init__(self, entity_id: int, x: int, y: int, z: int, stage: int):
        super().__init__()

        self.entity_id = entity_id
        self.x, self.y, self.z = x, y, z
        self.stage = stage

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_varint(self.entity_id)
            .write_position(self.x, self.y, self.z)
            .write("b", self.stage)
        )


class PlayAnimationServerBound(ServerBoundPacket):
    """Sent when a client's arm swings. (Client -> Server)

    :param int hand: Either main hand (0) or offhand (1).
    :ivar int id: Unique packet ID.
    :ivar hand:
    """

    id = 0x2C

    def __init__(self, hand: int):
        super().__init__()

        self.hand = hand

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayAnimationServerBound:
        return cls(buf.read_varint())


class PlayOpenBook(ClientBoundPacket):
    """Sent when a player right clicks a signed book. (Server -> Client)

    :param int hand: The hand used, either main (0) or offhand (1).
    :ivar int id: Unique packet ID.
    :ivar hand:
    """

    id = 0x2D

    def __init__(self, hand: int):
        super().__init__()

        self.hand = hand

    def pack(self) -> Buffer:
        return Buffer().write_varint(self.hand)



"""
Contains a "strict" implementation of abc.ABC, as only Python 3.10+ actually enforces abc.ABC.

- Subclasses of abstract classes must implement all abstract methods/classmethods/staticmethods
- A method's typehints are also enforced, if abstract method a has a return annotation of MyClass,
    the implemented method must have a return annotation of MyClass or a subclass of MyClass.
"""

from typing import Optional, Tuple, Type

__all__ = ("abstract", "StrictABC")


class AbstractObjData:
    __slots__ = ("optional", "annotations")

    def __init__(self, optional: bool = False, annotations: dict = None):
        self.optional = optional
        self.annotations = annotations


def abstract(obj):
    obj.__abstract__ = AbstractObjData(annotations=getattr(obj, "__annotations__", None))
    return obj


def optionalabstract(obj):
    obj.__abstract__ = AbstractObjData(
        optional=True, annotations=getattr(obj, "__annotations__", None)
    )
    return obj


def is_abstract(obj) -> bool:
    return hasattr(obj, "__abstract__")


def check_annotations(a: dict, b: dict) -> bool:
    for k, v in a.items():
        if k not in b:
            return False

        if type(v) is str:
            return type(b[k]) is str and v == b[k]

        try:
            if not issubclass(b[k], v):
                return False
        except TypeError:
            if b[k] is not v:
                return False

    return True


class UnimplementedAbstractError(Exception):
    def __init__(self, cls_name: str, obj_name: str):
        super().__init__(f"Class {cls_name} didn't implement abstract {obj_name}.")


class MismatchedOverridenAbstractError(Exception):
    def __init__(
        self, cls_name: str, obj_name: str, abstract_annotations: dict, obj_annotations: dict
    ):
        super().__init__(
            f"Overriden abstract {obj_name} of {cls_name} has mismatched annotations: {obj_name}({self.fmt_annos(abstract_annotations)}) != {obj_name}({self.fmt_annos(obj_annotations)})."
        )

    @staticmethod
    def fmt_annos(dct: dict) -> str:
        return ",".join(f"{k}:{getattr(v, '__name__', str(v))}" for k, v in dct.items())


class StrictABCMeta(type):
    def __new__(cls, name: str, bases: Tuple[Type[object]], dct: dict):
        self = super().__new__(cls, name, bases, dct)

        # iterate through base classes
        for base in bases:
            # iterate through objects
            for obj_name in dir(base):
                obj = getattr(base, obj_name)

                abstract_data: Optional[AbstractObjData] = getattr(obj, "__abstract__", None)

                self_obj = getattr(self, obj_name, None)

                # check if it's an abstract object
                if abstract_data is not None:
                    # check to see if this class implements that abstract object
                    if hasattr(self_obj, "__abstract__") and not abstract_data.optional:
                        raise UnimplementedAbstractError(self.__name__, obj.__name__)

                    if type(self_obj) is not type(obj) and not abstract_data.optional:
                        raise UnimplementedAbstractError(self.__name__, obj.__name__)

                    # check annotations if there are any
                    if abstract_data.annotations is not None:
                        self_obj_annotations = getattr(self_obj, "__annotations__", None)

                        if self_obj_annotations is None and abstract_data.annotations is not None:
                            raise MismatchedOverridenAbstractError(
                                self.__name__,
                                obj.__name__,
                                abstract_data.annotations,
                                self_obj_annotations,
                            )

                        if not check_annotations(abstract_data.annotations, self_obj_annotations):
                            raise MismatchedOverridenAbstractError(
                                self.__name__,
                                obj.__name__,
                                abstract_data.annotations,
                                self_obj_annotations,
                            )

        return self


class StrictABC(metaclass=StrictABCMeta):
    pass


# Example:
# class AbstractTest(StrictABC):
#     def __init__(self, something):
#         self.something = something

# class AbstractTest2(AbstractTest):
#     @abstract
#     def my_abstract_method(self, a: str):
#         print("in abstract method")

#     @classmethod
#     @abstract
#     def my_abstract_classmethod(self, a: str):
#         print("in abstract classmethod")

# class Test(AbstractTest2):
#     def my_abstract_method(self, a: str):
#         print("in method")

#     @classmethod
#     def my_abstract_classmethod(self, a: str):
#         print("in classmethod")



import asyncio
from typing import Union

from pymine_net.net.asyncio.stream import AsyncTCPStream
from pymine_net.net.client import AbstractTCPClient
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket
from pymine_net.types.packet_map import PacketMap

__all__ = ("AsyncTCPClient",)


class AsyncTCPClient(AbstractTCPClient):
    """An async connection over a TCP socket for reading + writing Minecraft packets."""

    def __init__(self, host: str, port: int, protocol: Union[int, str], packet_map: PacketMap):
        super().__init__(host, port, protocol, packet_map)

        self.stream: AsyncTCPStream = None

    async def connect(self) -> None:
        _, writer = await asyncio.open_connection(self.host, self.port)
        self.stream = AsyncTCPStream(writer)

    async def close(self) -> None:
        self.stream.close()
        await self.stream.wait_closed()

    async def read_packet(self) -> ClientBoundPacket:
        packet_length = await self.stream.read_varint()
        return self._decode_packet(await self.stream.readexactly(packet_length))

    async def write_packet(self, packet: ServerBoundPacket) -> None:
        self.stream.write(self._encode_packet(packet))
        await self.stream.drain()



from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ServerBoundPacket

__all__ = ("HandshakeHandshake",)


class HandshakeHandshake(ServerBoundPacket):
    """Initiates the connection between the server and client. (Client -> Server)

    :param int protocol: Protocol version to be used.
    :param str address: The host/address the client is connecting to.
    :param int port: The port the client is connection on.
    :param int next_state: The next state which the server should transfer to. 1 for status, 2 for login.
    :ivar int id: Unique packet ID.
    :ivar protocol:
    :ivar address:
    :ivar port:
    :ivar next_state:
    """

    id = 0x00

    def __init__(self, protocol: int, address: str, port: int, next_state: int):
        super().__init__()

        self.protocol = protocol
        self.address = address
        self.port = port
        self.next_state = next_state

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_varint(self.protocol)
            .write_string(self.address)
            .write("H", self.port)
            .write_varint(self.next_state)
        )

    @classmethod
    def unpack(cls, buf: Buffer) -> HandshakeHandshake:
        return cls(buf.read_varint(), buf.read_string(), buf.read("H"), buf.read_varint())



"""Contains packets that are related to particles."""

from __future__ import annotations

from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket

__all__ = ("PlayParticle",)


class PlayParticle(ClientBoundPacket):
    """Sent by server to make the client display particles. (Server -> Client)

    :param int particle_id: ID of the particle.
    :param bool long_distance: If true, particle distance increases to 65536 from 256.
    :param int x: X coordinate of the particle.
    :param int y: Y coordinate of the particle.
    :param int z: Z coordinate of the particle.
    :param float particle_data: Particle data.
    :param int particle_count: How many particles to display.
    :param dict data: More particle data.
    :ivar int id: Unique packet ID.
    :ivar particle_id:
    :ivar long_distance:
    :ivar x:
    :ivar y:
    :ivar z:
    :ivar particle_data:
    :ivar particle_count:
    :ivar data:
    """

    id = 0x24

    def __init__(
        self,
        particle_id: int,
        long_distance: bool,
        x: float,
        y: float,
        z: float,
        offset_x: float,
        offset_y: float,
        offset_z: float,
        particle_data: float,
        particle_count: int,
        data: dict,
    ):
        super().__init__()

        self.part_id = particle_id
        self.long_dist = long_distance
        self.x = x
        self.y = y
        self.z = z
        self.offset_x = offset_x
        self.offset_y = offset_y
        self.offset_z = offset_z
        self.particle_data = particle_data
        self.particle_count = particle_count
        self.data = data

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write("i", self.part_id)
            .write("?", self.long_dist)
            .write("d", self.x)
            .write("d", self.y)
            .write("d", self.z)
            .write("f", self.offset_x)
            .write("f", self.offset_y)
            .write("f", self.offset_z)
            .write("f", self.particle_data)
            .write("i", self.particle_count)
            .write_particle(self.data)
        )



import asyncio
from typing import Dict, Tuple, Union

from pymine_net.net.asyncio.stream import AsyncTCPStream
from pymine_net.net.server import AbstractProtocolServer, AbstractProtocolServerClient
from pymine_net.strict_abc import abstract
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket
from pymine_net.types.packet_map import PacketMap

__all__ = ("AsyncProtocolServerClient", "AsyncProtocolServer")


class AsyncProtocolServerClient(AbstractProtocolServerClient):
    def __init__(self, stream: AsyncTCPStream, packet_map: PacketMap):
        super().__init__(stream, packet_map)
        self.stream = stream  # redefine this cause typehints

    async def read_packet(self) -> ServerBoundPacket:
        length = await self.stream.read_varint()
        return self._decode_packet(await self.stream.readexactly(length))

    async def write_packet(self, packet: ClientBoundPacket) -> None:
        self.stream.write(self._encode_packet(packet))
        await self.stream.drain()


class AsyncProtocolServer(AbstractProtocolServer):
    def __init__(self, host: str, port: int, protocol: Union[int, str], packet_map: PacketMap):
        super().__init__(host, port, protocol, packet_map)

        self.connected_clients: Dict[Tuple[str, int], AsyncProtocolServerClient] = {}

        self.server: asyncio.AbstractServer = None

    async def run(self) -> None:
        self.server = await asyncio.start_server(self._client_connected_cb, self.host, self.port)

    async def close(self) -> None:
        self.server.close()
        await self.server.wait_closed()

    async def _client_connected_cb(
        self, _: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        client = AsyncProtocolServerClient(AsyncTCPStream(writer), self.packet_map)

        self.connected_clients[client.stream.remote] = client

        await self.new_client_connected(client)

    @abstract
    async def new_client_connected(self, client: AsyncProtocolServerClient) -> None:
        pass



"""Contains packets related to blocks."""

from __future__ import annotations

from typing import List, Tuple

import pymine_net.types.nbt as nbt
from pymine_net.types.buffer import Buffer
from pymine_net.types.packet import ClientBoundPacket, ServerBoundPacket

__all__ = (
    "PlayBlockAction",
    "PlayBlockChange",
    "PlayQueryBlockNBT",
    "PlayPlayerBlockPlacement",
    "PlayNBTQueryResponse",
    "PlayMultiBlockChange",
)


class PlayBlockAction(ClientBoundPacket):
    """This packet is used for a number of actions and animations performed by blocks. (Server -> Client)

    :param int x: The x coordinate of the location where this occurs.
    :param int y: The y coordinate of the location where this occurs.
    :param int z: The z coordinate of the location where this occurs.
    :param int action_id: Block action ID, see here: https://wiki.vg/Block_Actions.
    :param int action_param: Action param of the action, see here: https://wiki.vg/Block_Actions.
    :param int block_type: The type of block which the action is for.
    :ivar int id: Unique packet ID.
    :ivar x:
    :ivar y:
    :ivar z:
    :ivar action_id:
    :ivar action_param:
    :ivar block_type:
    """

    id = 0x0B

    def __init__(self, x: int, y: int, z: int, action_id: int, action_param: int, block_type: int):
        super().__init__()

        self.x, self.y, self.z = x, y, z
        self.action_id = action_id
        self.action_param = action_param
        self.block_type = block_type

    def pack(self) -> Buffer:
        return (
            Buffer()
            .write_position(self.x, self.y, self.z)
            .write("B", self.action_id)
            .write("B", self.action_param)
            .write_varint(self.block_type)
        )


class PlayBlockChange(ClientBoundPacket):
    """Fired when a block is changed within the render distance. (Server -> Client)

    :param int x: The x coordinate of the location where this occurs.
    :param int y: The y coordinate of the location where this occurs.
    :param int z: The z coordinate of the location where this occurs.
    :param int block_id: Block ID of what to change the block to.
    :ivar int id: Unique packet ID.
    :ivar x:
    :ivar y:
    :ivar z:
    :ivar block_id:
    """

    id = 0x0C

    def __init__(self, x: int, y: int, z: int, block_id: int):
        super().__init__()

        self.x, self.y, self.z = x, y, z
        self.block_id = block_id

    def pack(self) -> Buffer:
        return Buffer().write_position(self.x, self.y, self.z).write_varint(self.block_id)


class PlayQueryBlockNBT(ServerBoundPacket):
    """Used when SHIFT+F3+I is used on a block. (Client -> Server)

    :param int transaction_id: Number used to verify that a response matches.
    :param int x: The x coordinate of the block.
    :param int y: The y coordinate of the block.
    :param int z: The z coordinate of the block.
    :ivar int id: Unique packet ID.
    :ivar transaction_id:
    :ivar x:
    :ivar y:
    :ivar z:
    """

    id = 0x01

    def __init__(self, transaction_id: int, x: int, y: int, z: int):
        super().__init__()

        self.transaction_id = transaction_id
        self.x, self.y, self.z = x, y, z

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayQueryBlockNBT:
        return cls(buf.read_varint(), *buf.read_position())


class PlayPlayerBlockPlacement(ServerBoundPacket):
    """Sent by the client when it places a block. (Client -> Server)

    :param int hand: The hand used, either main hand (0), or offhand (1).
    :param int x: The x coordinate of the block.
    :param int y: The y coordinate of the block.
    :param int z: The z coordinate of the block.
    :param int face: The face of the block, see here: https://wiki.vg/Protocol#Player_Block_Placement.
    :param float cur_pos_x: The x position of the crosshair on the block.
    :param float cur_pos_y: The y position of the crosshair on the block.
    :param float cur_pos_z: The z position of the crosshair on the block.
    :param bool inside_block: True if the player's head is inside the block.
    :ivar int id: Unique packet ID.
    :ivar hand:
    :ivar x:
    :ivar y:
    :ivar z:
    :ivar face:
    :ivar cur_pos_x:
    :ivar cur_pos_y:
    :ivar cur_pos_z:
    :ivar inside_block:
    """

    id = 0x2E

    def __init__(
        self,
        hand: int,
        x: int,
        y: int,
        z: int,
        face: int,
        cur_pos_x: float,
        cur_pos_y: float,
        cur_pos_z: float,
        inside_block: bool,
    ):
        super().__init__()

        self.hand = hand
        self.x, self.y, self.z = x, y, z
        self.face = face
        self.cur_pos_x = cur_pos_x
        self.cur_pos_y = cur_pos_y
        self.cur_pos_z = cur_pos_z
        self.inside_block = inside_block

    @classmethod
    def unpack(cls, buf: Buffer) -> PlayPlayerBlockPlacement:
        return cls(
            buf.read_varint(),
            *buf.read_position(),
            buf.read_varint(),
            buf.read("f"),
            buf.read("f"),
            buf.read("f"),
            buf.read("?"),
        )


class PlayNBTQueryResponse(ClientBoundPacket):
    """Sent by the client when it places a block. (Server -> Client)

    :param int transaction_id:
    :param nbt.TAG nbt_tag:
    :ivar int id: Unique packet ID.
    :ivar transaction_id:
    :ivar nbt_tag:
    """

    id = 0x60

    def __init__(self, transaction_id: int, nbt_tag: nbt.TAG):
        super().__init__()

        self.transaction_id = transaction_id
        self.nbt_tag = nbt_tag

    def pack(self) -> Buffer:
        return Buffer().write_varint(self.transaction_id).write_nbt(self.nbt_tag)


class PlayMultiBlockChange(ClientBoundPacket):
    """Sent whenever 2 or more blocks change in the same chunk on the same tick. (Server -> Client)

    :param int chunk_sect_x: The x coordinate of the chunk section.
    :param int chunk_sect_y: The y coordinate of the chunk section.
    :param int chunk_sect_z: The z coordinate of the chunk section.
    :param bool trust_edges: The inverse of preceding PlayUpdateLight packet's trust_edges bool
    :param list blocks: The changed blocks, formatted like [block_id, local_x, local_y, local_z].
    :ivar int id: Unique packet ID.
    :ivar chunk_sect_x:
    :ivar chunk_sect_y:
    :ivar chunk_sect_z:
    :ivar trust_edges:
    :ivar blocks:
    """

    id = 0x3F

    def __init__(
        self,
        chunk_sect_x: int,
        chunk_sect_y: int,
        chunk_sect_z: int,
        trust_edges: bool,
        blocks: List[Tuple[int, int, int, int]],
    ):
        super().__init__()

        self.chunk_sect_x = chunk_sect_x
        self.chunk_sect_y = chunk_sect_y
        self.chunk_sect_z = chunk_sect_z
        self.trust_edges = trust_edges
        self.blocks = blocks

    def pack(self) -> Buffer:
        buf = (
            Buffer()
            .write_varint(
                ((self.chunk_sect_x & 0x3FFFFF) << 42)
                | (self.chunk_sect_y & 0xFFFFF)
                | ((self.chunk_sect_z & 0x3FFFFF) << 20)
            )
            .write("?", self.trust_edges)
            .write_varint(len(self.blocks))
        )

        for block_id, local_x, local_y, local_z in self.blocks:
            buf.write_varint(block_id << 12 | (local_x << 8 | local_z << 4 | local_y))

        return buf



from typing import Union

__all__ = ("Registry",)


class Registry:
    def __init__(
        self,
        data: Union[dict, list, tuple],
        data_reversed: Union[dict, list, tuple] = None,
    ):
        self.data_reversed = data_reversed

        if isinstance(data, dict):
            self.data = data

            if data_reversed is None:
                self.data_reversed = {v: k for k, v in data.items()}
        elif isinstance(data, (list, tuple)):
            self.data_reversed = data
            self.data = {v: i for i, v in enumerate(self.data_reversed)}
        else:
            raise TypeError(
                "Creating a registry from something other than a dict, tuple, or list isn't supported"
            )

    def encode(self, key: object) -> object:
        """Key -> value, most likely an identifier to an integer."""

        return self.data[key]

    def decode(self, value: object) -> object:
        """Value -> key, most likely a numeric id to a string identifier."""

        return self.data_reversed[value]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pymine_net',
 'pymine_net.net',
 'pymine_net.net.asyncio',
 'pymine_net.net.socket',
 'pymine_net.packets',
 'pymine_net.packets.v_1_18_1.handshaking',
 'pymine_net.packets.v_1_18_1.login',
 'pymine_net.packets.v_1_18_1.play',
 'pymine_net.packets.v_1_18_1.status',
 'pymine_net.types']

package_data = \
{'': ['*']}

install_requires = \
['cryptography>=36.0.1,<37.0.0', 'mutf8>=1.0.6,<2.0.0']

setup_kwargs = {
    'name': 'pymine-net',
    'version': '0.2.0',
    'description': 'Networking library for Minecraft in Python',
    'long_description': "# PyMine-Net\n[![CodeFactor](https://www.codefactor.io/repository/github/py-mine/pymine-net/badge)](https://www.codefactor.io/repository/github/py-mine/pymine-net)\n![code size](https://img.shields.io/github/languages/code-size/py-mine/PyMine-Net?color=0FAE6E)\n![code style](https://img.shields.io/badge/code%20style-black-000000.svg)\n\n*PyMine-Net - an extensible and modular Minecraft networking library in Python*\n\n## Features\n- Buffer class with methods for dealing with various types and data formats used by Minecraft\n- High level abstractions for packets with a system to seperate different protocol's packets\n- Miscellaneous classes for dealing with different Minecraft data structures relevant to networking\n",
    'author': 'PyMine-Net Developers & Contributors',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/py-mine/pymine-net',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from pymine_net.strict_abc import StrictABC, abstract, optionalabstract
from pymine_net.types.buffer import Buffer

__all__ = ("Packet", "ServerBoundPacket", "ClientBoundPacket")


class Packet(StrictABC):
    """Base Packet class.

    :param int id: Packet identification number. Defaults to None.
    :ivar id:
    """

    def __init__(self):
        self.id: int = self.__class__.id


class ServerBoundPacket(Packet):
    """Base Packet class for packets received from the client. (Client -> Server)

    :param int id: Packet identification number. Defaults to None.
    :ivar id:
    """

    @optionalabstract
    def pack(self) -> Buffer:
        raise NotImplementedError

    @abstract
    @classmethod
    def unpack(cls, buf: Buffer) -> ServerBoundPacket:
        pass


class ClientBoundPacket(Packet):
    """Base Packet class for packets to be sent from the server. (Server -> Client)

    :param int id: Packet identification number. Defaults to None.
    :ivar id:
    """

    @abstract
    def pack(self) -> Buffer:
        pass

    @classmethod
    @optionalabstract
    def unpack(cls, buf: Buffer) -> ClientBoundPacket:
        raise NotImplementedError



"""Contains packets related to the in-game map item."""

from __future__ import annotations

from typing import List, Optional, Tuple

from pymine_net.types.buffer import Buffer
from pymine_net.types.chat import Chat
from pymine_net.types.packet import ClientBoundPacket

__all__ = ("PlayMapData",)


class PlayMapData(ClientBoundPacket):
    """Updates a rectangular area on a map item. (Server -> Client)

    :param int map_id: ID of the map to be modified.
    :param int scale: Zoom of the map (0 fully zoomed in - 4 fully zoomed out).
    :param bool locked: Whether the map has been locked in a cartography table or not.
    :param bool tracking_pos: Whether the player and other icons are shown on the map.
    :param List[Tuple[int, int, int, int, bool, Optional[Chat]]] icons: List of icons shown on the map.
    :param int columns: Number of columns being updated.
    :param Optional[int] rows: Number of rows being updated, only if columns > 0.
    :param Optional[int] x: X offset of the westernmost column, only if columns > 0.
    :param Optional[int] z: Z offset of the northernmost row, only if columns > 0.
    :param bytes data: The map data, see https://minecraft.fandom.com/wiki/Map_item_format.
    :ivar int id:
    :ivar map_id:
    :ivar scale:
    :ivar locked:
    :ivar tracking_pos:
    :ivar icons:
    :ivar columns:
    :ivar rows:
    :ivar x:
    :ivar z:
    :ivar data:
    """

    id = 0x27

    def __init__(
        self,
        map_id: int,
        scale: int,
        locked: bool,
        tracking_pos: bool,
        icons: List[Tuple[int, int, int, int, bool, Optional[Chat]]],
        columns: int,
        rows: int = None,
        x: int = None,
        z: int = None,
        data: bytes = None,
    ):
        super().__init__()

        self.map_id = map_id
        self.scale = scale
        self.tracking_pos = tracking_pos
        self.locked = locked
        self.icons = icons
        self.columns = columns
        self.rows = rows
        self.x = x
        self.z = z
        self.data = data

    def pack(self) -> Buffer:
        buf = (
            Buffer()
            .write_varint(self.map_id)
            .write_byte(self.scale)
            .write("?", self.locked)
            .write("?", self.tracking_pos)
            .write_varint(len(self.icons))
        )

        for (icon_type, x, z, direction, display_name) in self.icons:
            (
                buf.write_varint(icon_type)
                .write_byte(x)
                .write_byte(z)
                .write_byte(direction)
                .write_optional(buf.write_chat, display_name)
            )

        buf.write("B", self.columns)

        if len(self.columns) < 1:
            return buf

        return (
            buf.write("B", self.rows)
            .write_byte(self.x)
            .write_byte(self.y)
            .write_varint(len(self.data))
            .write_bytes(self.data)
        )
